#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import logging
import time
from typing import Dict, List, Tuple
from klayout import db
import os

from .writer import Writer, remap_layers
from ..layout import layers
from ..lef import types as lef

logger = logging.getLogger(__name__)


def _decompose_region(region: db.Region, ignore_non_rectilinear: bool = False) -> List[db.Box]:
    """
    Decompose a `db.Region` of multiple `db.Polygon`s into non-overlapping rectangles (`db.Box`).
    :param region:
    :param ignore_non_rectilinear: If set to `True` then non-rectilinear polygons are skipped.
    :return: Returns the list of rectangles.
    """
    trapezoids = region.decompose_trapezoids_to_region()
    logger.debug("Number of trapezoids: {}".format(trapezoids.size()))
    rectangles = []
    for polygon in trapezoids.each():
        box = polygon.bbox()

        if db.Polygon(box) != polygon:
            msg = "Cannot decompose into rectangles. Something is not rectilinear!"
            if not ignore_non_rectilinear:
                logger.error(msg)
                assert False, msg
            else:
                logger.warning(msg)

        rectangles.append(box)
    return rectangles


def generate_lef_macro(cell_name: str,
                       pin_geometries: Dict[str, List[Tuple[str, db.Shape]]],
                       pin_direction: Dict[str, lef.Direction],
                       pin_use: Dict[str, lef.Use],
                       scaling_factor: float = 1,
                       use_rectangles_only: bool = False,
                       ) -> lef.Macro:
    """
    Assemble a LEF MACRO structure containing the pin shapes.
    :param cell_name: Name of the cell as it will appear in the LEF file.
    :param pin_geometries: A dictionary mapping pin names to geometries: Dict[pin name, List[(layer name, klayout Shape)]]
    :param pin_direction:
    :param pin_use:
    :param use_rectangles_only: Decompose all polygons into rectangles. Non-rectilinear shapes are dropped.
    :return: Returns a `lef.Macro` object containing the pin information of the cell.

    # TODO: FOREIGN statement (reference to GDS)
    """

    logger.debug("Generate LEF MACRO structure for {}.".format(cell_name))
    logger.debug(f"Scaling factor = {scaling_factor}.")

    f = scaling_factor

    pins = []
    # Create LEF Pin objects containing geometry information of the pins.
    for pin_name, ports in pin_geometries.items():

        layers = []

        for layer_name, shape in ports:
            # Convert all non-regions into a region
            region = db.Region()
            region.insert(shape)
            region.merge()
            if use_rectangles_only:
                # Decompose into rectangles.
                boxes = _decompose_region(region)
                region = db.Region()
                region.insert(boxes)

            geometries = []
            for p in region.each():
                polygon = p.to_simple_polygon()

                box = polygon.bbox()
                is_box = db.SimplePolygon(box) == polygon

                if is_box:
                    rect = lef.Rect((box.p1.x * f, box.p1.y * f), (box.p2.x * f, box.p2.y * f))
                    geometries.append(rect)
                else:
                    # Port is a polygon
                    # Convert `db.Point`s into LEF points.
                    points = [(p.x * f, p.y * f) for p in polygon.each_point()]
                    poly = lef.Polygon(points)
                    geometries.append(poly)

            layers.append((lef.Layer(layer_name), geometries))

        port = lef.Port(CLASS=lef.Class.CORE,
                        geometries=layers)

        # if pin_name not in pin_direction:
        #     msg = "I/O direction of pin '{}' is not defined.".format(pin_name)
        #     logger.error(msg)
        #     assert False, msg
        #
        # if pin_name not in pin_use:
        #     msg = "Use of pin '{}' is not defined. Must be one of (CLK, SIGNAL, POWER, ...)".format(pin_name)
        #     logger.error(msg)
        #     assert False, msg

        pin = lef.Pin(pin_name=pin_name,
                      direction=lef.Direction.INOUT,  # TODO: find direction
                      use=lef.Use.SIGNAL,  # TODO: correct use
                      shape=lef.Shape.ABUTMENT,
                      port=port,
                      property={},
                      )
        pins.append(pin)

    macro = lef.Macro(
        name=cell_name,
        macro_class=lef.MacroClass.CORE,
        foreign=lef.Foreign(cell_name, lef.Point(0, 0)),
        origin=lef.Point(0, 0),
        symmetry={lef.Symmetry.X, lef.Symmetry.Y, lef.Symmetry.R90},
        site="CORE",
        pins=pins,
        obstructions=[]
    )

    return macro


class LefWriter(Writer):

    def __init__(self,
                 output_map: Dict[str, Tuple[int, int]],
                 db_unit: float = 1e-6,
                 use_rectangles_only: bool = False):
        """

        :param output_map:
        :param db_unit: Database unit in meters. Default is 1um (1e-6 m)
        :param use_rectangles_only: Convert all polygons into rectangles. Non-rectilinear shapes are dropped.
        """
        self.db_unit = db_unit
        self.output_map = output_map
        self.scaling_factor = 1
        self.use_rectangles_only = use_rectangles_only

    def write_layout(self,
                     layout: db.Layout,
                     pin_geometries: Dict[str, List[Tuple[str, db.Shape]]],
                     top_cell: db.Cell,
                     output_dir: str,
                     ) -> None:
        # Re-map layers
        layout = remap_layers(layout, self.output_map)

        # Compute correct scaling factor.
        # klayout expects dbu to be in µm, the tech file takes it in meters.
        logger.debug(f"LEF db_unit = {self.db_unit} [m]")
        scaling_factor = self.db_unit / (layout.dbu)
        scaling_factor *= self.scaling_factor  # Allow to make corrections from the tech file.

        # Write LEF
        # Create and populate LEF Macro data structure.
        # TODO: pass correct USE and DIRECTION
        lef_macro = generate_lef_macro(top_cell.name,
                                       pin_geometries=pin_geometries,
                                       pin_use=None,
                                       pin_direction=None,
                                       scaling_factor=scaling_factor,
                                       use_rectangles_only=self.use_rectangles_only)

        # Write LEF
        lef_file_name = "{}.lef".format(top_cell.name)
        lef_output_path = os.path.join(output_dir, lef_file_name)

        with open(lef_output_path, "w") as f:
            logger.info("Write LEF: {}".format(lef_output_path))
            f.write(lef.lef_format(lef_macro))



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import sys

import networkx as nx
from typing import Tuple, Iterable

# klayout.db should not be imported if script is run from KLayout GUI.
if 'pya' not in sys.modules:
    import klayout.db as pya


def is_inside(p: Tuple[int, int], r: pya.Region, margin: int):
    """ Test if point is inside region with some margin.
    """
    p = pya.Point(*p)
    v = pya.Vector(margin, margin)
    box = pya.Region(pya.Box(p - v, p + v))
    return not box.inside(r).is_empty()


def interacts(p: Tuple[int, int], r: pya.Region, margin: int):
    """ Test if point is close (margin) to region.
    """
    p = pya.Point(*p)
    v = pya.Vector(margin, margin)
    box = pya.Region(pya.Box(p - v, p + v))
    return not r.interacting(box).is_empty()


def is_edge_inside(graph_edge, r: pya.Region, margin: int):
    """Checks if the edge of the graph lies inside the region when mapped to a layout.
    """

    assert margin > 0, "Path with width 0 is considered non-existent."

    (l1, (x1, y1)), (l2, (x2, y2)) = graph_edge

    p = pya.Region(pya.Path([pya.Point(x1, y1), pya.Point(x2, y2)], margin, margin, margin))

    return not p.inside(r).is_empty()


def inside(points: Iterable[Tuple[int, int]], r: pya.Region, margin: int):
    """ Get all points that lie inside the region.
    """
    return [p for p in points if is_inside(p, r, margin)]


def interacting(points: Iterable[Tuple[int, int]], r: pya.Region, margin: int):
    """ Get all points that are close (margin) to the region.
    """
    return [p for p in points if interacts(p, r, margin)]


def edges_inside(g: nx.Graph, r: pya.Region, margin: int):
    """ Get all graph edges that lie inside the region.
    """
    return [e for e in g.edges if is_edge_inside(e, r, margin)]



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import klayout.db as pya


def fill_notches(region: pya.Region, minimum_notch: int) -> pya.Region:
    """ Fill notches in a pya.Region.
    :param region:
    :param minimum_notch:
    :return:
    """

    notches = region.notch_check(minimum_notch)
    spaces = region.space_check(minimum_notch)
    notches = list(notches) + list(spaces)
    s = pya.Shapes()
    s.insert(region)
    for edge_pair in notches:
        a, b = edge_pair.first, edge_pair.second
        # Find smaller edge (a)
        a, b = sorted((a, b), key=lambda e: e.length())

        # Construct a minimal box to fill the notch
        box = a.bbox()
        # Extend box of shorted edge by points of longer edge
        box1 = box + b.p1
        box2 = box + b.p2

        # Take the smaller box.
        min_box = min([box1, box2], key=lambda b: b.area())

        s.insert(min_box)

    result = pya.Region(s)
    result.merge()
    return result




#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx
import numpy as np
from .graphrouter import GraphRouter
from .signal_router import SignalRouter
from .multi_via_router import MultiViaRouter

from itertools import chain, count, combinations, product
from collections import Counter

from typing import Any, Dict, List, AbstractSet, Optional

import logging

logger = logging.getLogger(__name__)


class PathFinderGraphRouter(GraphRouter):

    def __init__(self,
                 detail_router: SignalRouter,
                 is_virtual_edge_fn=None):
        self.detail_router = detail_router

    def route(self,
              graph: nx.Graph,
              signals: Dict[Any, List[Any]],
              reserved_nodes: Optional[Dict] = None,
              node_conflict: Optional[Dict[Any, AbstractSet[Any]]] = None,
              equivalent_nodes: Optional[Dict[Any, AbstractSet[Any]]] = None,
              is_virtual_node_fn=None
              ) -> Dict[Any, nx.Graph]:
        return _route(self.detail_router,
                      graph,
                      signals=signals,
                      reserved_nodes=reserved_nodes,
                      node_conflict=node_conflict,
                      equivalent_nodes=equivalent_nodes,
                      is_virtual_node_fn=is_virtual_node_fn)


def _compute_tree_weight(routing_tree: nx.Graph, edge_cost_fn, is_virtual_edge_fn) -> float:
    """
    Compute the weight of a routing tree based on a edge cost function.
    :param routing_tree: A tree graph.
    :param edge_cost_fn: edge_cost((a,b)) = 'cost of edge from node a to node b'
    :param is_virtual_edge_fn: Tells wether an edge is 'virtual'. Edges adjacent to virtual nodes are not accounted in the tree weight.
    :return: The tree weight
    """
    tree_weights = sum(edge_cost_fn(e) for e in routing_tree.edges() if not is_virtual_edge_fn(e))
    return tree_weights


def _compute_tree_weights(routing_trees: Dict[Any, nx.Graph], edge_cost_fn, is_virtual_edge_fn) -> Dict[Any, float]:
    """
    Helper function which calls `_compute_tree_weight` for each element of a dictionary.
    :param routing_trees:
    :param edge_cost_fn:
    :param is_virtual_edge_fn:
    :return:
    """
    tree_weights = {signal_name: _compute_tree_weight(rt, edge_cost_fn, is_virtual_edge_fn)
                    for signal_name, rt in routing_trees.items()}
    return tree_weights


def _route(detail_router: SignalRouter,
           graph: nx.Graph,
           signals: Dict[Any, List[Any]],
           reserved_nodes: Optional[Dict] = None,
           node_conflict: Optional[Dict[Any, AbstractSet[Any]]] = None,
           equivalent_nodes: Optional[Dict[Any, AbstractSet[Any]]] = None,
           is_virtual_node_fn=None) -> Dict[Any, nx.Graph]:
    """ Route multiple signals in the graph.
    Based on PathFinder algorithm.

    Parameters
    ----------
    :param detail_router: Underlying detail router.
    :param graph               :       networkx.Graph
                            Graph representing the routing grid.
    :param signals  :       Dict[node name, List[node]]
                            Signals to be routed. Each signal is represented by its terminal nodes.
    :param reserved_nodes: An optional dict which specifies nodes that are reserved for a specific net.
    Dict[net_name, set of nodes].
    :param node_conflict: Dict[node, Set[node]]
    Tells which other nodes are blocked by a node. A node might block its direct neigbhours to ensure minimum spacing.

    :returns : A list of `networkx.Graph`s representing the routes of each signal.
    """

    assert isinstance(signals, dict)

    logger.info('Start global routing.')

    assert nx.is_connected(graph), Exception("Cannot route in unconnected graph.")

    if reserved_nodes is None:
        reserved_nodes = dict()

    if node_conflict is None:
        node_conflict = dict()

    if is_virtual_node_fn is None:
        def is_virtual_node_fn(_):
            return False

    # Costs
    default_edge_cost = 1
    default_node_cost = 0.1
    edge_base_cost = {}  # Base cost of resource. Depends on layer type, geometries, ... (bn)

    node_base_cost = {}  # Base cost of resource. Depends on layer type, geometries, ... (bn)
    node_history_cost = {}  # Related to past congestion. (hn)
    node_present_sharing_cost = {}  # Related to number of signals using this resource in the current iteration. (pn)

    # Import edge costs from graph G
    for a, b, data in graph.edges(data=True):
        assert 'weight' in data, Exception('Edge has no weight: ', (a, b))
        w = data['weight']
        edge_base_cost[(a, b)] = w
        edge_base_cost[(b, a)] = w

        # if a not in node_base_cost:
        #     node_base_cost[a] = 0
        #
        # if b not in node_base_cost:
        #     node_base_cost[b] = 0
        #
        # node_base_cost[a] += w//2
        # node_base_cost[b] += w//2

    def is_virtual_edge(e) -> bool:
        a, b = e
        return is_virtual_node_fn(a) or is_virtual_node_fn(b)

    # Pre-scaling: Normalize edge costs.
    edge_costs = [cost for edge, cost in edge_base_cost.items() if not is_virtual_edge(edge)]
    mean_edge_cost = np.mean(edge_costs)

    logger.debug('Mean edge cost (without virtual edges): {:.2f}'.format(mean_edge_cost))
    logger.debug('Pre-scaling factor for edge costs: 1/{:.2f}'.format(mean_edge_cost))

    edge_base_cost = {k: v / mean_edge_cost for k, v in edge_base_cost.items()}

    routing_trees = {name: nx.Graph() for name in signals.keys()}
    slack_ratios = {name: 1 for name in signals.keys()}

    # For each net create its own routing graph.
    # Some routing nodes are only available for a specific net.
    Gs = dict()
    all_reserved = set(chain(*reserved_nodes.values()))
    for net, terminals in signals.items():
        forbidden_nodes = all_reserved - set(reserved_nodes.get(net, {}))
        G2 = graph.copy()
        if forbidden_nodes:
            # Need to delete some nodes from G.

            G2.remove_nodes_from(forbidden_nodes)

            for t1, t2 in combinations(terminals, 2):
                assert nx.node_connectivity(G2, t1, t2) > 0, \
                    Exception("Graph has been disconnected by removal of reserved nodes ({}).".format(net))

            Gs[net] = G2

    # TODO: just use detail_router and let caller decide whether to use MultiViaRouter or not.
    multi_via_router = MultiViaRouter(detail_router, node_conflict)
    # multi_via_router = detail_router  # Don't use multi via router.

    history_cost_weight = 1
    node_present_sharing_cost_increment = 10

    max_iterations = 1000

    for j in count():

        if j >= max_iterations:
            raise Exception("Failed to route")

        logger.info('Routing iteration %d' % j)

        routing_order = sorted(signals.keys(), key=lambda i: (slack_ratios[i], i), reverse=True)
        logger.debug('Routing order: {}'.format(routing_order))
        node_present_sharing_cost.clear()

        for signal_name in routing_order:
            terminals = signals[signal_name]

            slack_ratio = slack_ratios[signal_name]

            def node_cost_fn(n):
                b = node_base_cost.get(n, 0)
                h = node_history_cost.get(n, 0) * history_cost_weight
                p = node_present_sharing_cost.get(n, 0) * node_present_sharing_cost_increment
                c = (1 + b + h) * (p + 1)
                return c * (1 - slack_ratio)

            def edge_cost_fn(e):
                (m, n) = e
                b = edge_base_cost[(m, n)]
                return b * slack_ratio

            st = multi_via_router.route(Gs.get(signal_name, graph), terminals, node_cost_fn, edge_cost_fn)

            routing_trees[signal_name] = st

            # Include colliding nodes.
            nodes = set(chain(*(node_conflict.get(n, {}) for n in st.nodes)))
            nodes.update(st.nodes)

            for n in st.nodes:
                for o in node_conflict.get(n, {n}):
                    if o not in node_present_sharing_cost:
                        node_present_sharing_cost[o] = 0
                    node_present_sharing_cost[o] += 1

        # Detect node collisions

        # Get all nodes that are used for routing.
        all_routing_tree_nodes = set(chain(*(tree.nodes
                                             for tree in routing_trees.values())))

        # For each signal, get a set of nodes that are used by it.
        signal_nodes = (set(chain(*(node_conflict.get(n, {n}) for n in tree.nodes)))
                        for tree in routing_trees.values())

        # Count how many times a node is used.
        node_sharing = Counter(chain(*signal_nodes))

        # Find collisions
        node_collisions = [k for k, v in node_sharing.items() if v > 1 and k in all_routing_tree_nodes]

        if equivalent_nodes:
            # Also mark nodes that are equivalent.
            node_collisions = set(chain(*(equivalent_nodes[n] for n in node_collisions)))

        has_collision = len(node_collisions) > 0

        # Compute weights of all signal trees.
        tree_weights = _compute_tree_weights(routing_trees, lambda e: edge_base_cost[e], is_virtual_edge)

        # Print routing tree weights.
        for signal_name, tree_weight in tree_weights.items():
            original_tree_weight = tree_weight * mean_edge_cost  # Invert pre-scaling.
            logger.debug('weight of {:>16}: {:.2f}'.format(signal_name, original_tree_weight))

        if not has_collision:
            logger.info("Global routing done in %d iterations", j)
            break

        # node_history_cost = {n: c*0.9 for n,c in node_history_cost.items()}

        # Increment history cost
        for n in node_collisions:
            node_history_cost[n] = node_history_cost.get(n, 0) + 1

        max_weight = max(tree_weights.values())
        slack_ratios = {n: t / max_weight for n, t in tree_weights.items()}
        slack_ratio_scaling = 0.1
        slack_ratios = {n: (s - 0.5) * slack_ratio_scaling + 0.5 for n, s in slack_ratios.items()}

    # Perform single-net optimizations.
    # For each net n: Rip up n while leaving the other nets untouched. Route n again.

    logger.info('Run single-net optimizations.')
    all_signal_names = signals.keys()
    for current_signal in all_signal_names:
        logger.debug('Single-net optimization: {}'.format(current_signal))
        other_signal_names = all_signal_names - current_signal

        other_routing_trees = {name: rt for name, rt in routing_trees.items() if name != current_signal}

        # Get the set of nodes occupied by the other signals.
        used_nodes = set(chain(*
                               (
                                   node_conflict.get(n, {n})
                                   for tree in other_routing_trees.values()
                                   for n in tree.nodes
                               )
                               )
                         )

        # Get the routing graph where all other signals are already routed.
        current_graph = Gs.get(current_signal, graph).copy()
        # Remove the nodes that are already in use by other signals.
        current_graph.remove_nodes_from(used_nodes)

        # Now the use the plain edge cost only (without history costs).
        def edge_cost_fn(e):
            (m, n) = e
            return edge_base_cost[(m, n)]

        # Find a route for the current signal.
        terminals = signals[current_signal]
        old_route = routing_trees[current_signal]
        new_route = multi_via_router.route(current_graph,
                                           terminals,
                                           node_cost_fn=lambda x: 0,
                                           edge_cost_fn=edge_cost_fn
                                           )

        old_weight = _compute_tree_weight(old_route, edge_cost_fn, is_virtual_edge)
        new_weight = _compute_tree_weight(new_route, edge_cost_fn, is_virtual_edge)
        logger.debug('Old weight for {}: {}'.format(current_signal, old_weight))
        logger.debug('New weight for {}: {}'.format(current_signal, new_weight))

        # Save the route.
        routing_trees[current_signal] = new_route

    return routing_trees


def test():
    logging.basicConfig(level=logging.INFO)
    import matplotlib.pyplot as plt
    from .signal_router import DijkstraRouter

    G = nx.Graph()

    num_x = 10
    num_y = 10
    x = range(0, num_x)
    y = range(0, num_y)

    # Store drawing positions of vertices. For plotting only.
    pos = {}

    # Construct mesh
    for name, (x, y) in enumerate(product(x, y)):
        G.add_node((x, y))
        pos[(x, y)] = (x, y)

        w = 1

        if x < num_x - 1 and not (1 <= y < 5 and x == 4):
            G.add_edge((x, y), (x + 1, y), weight=w, orientation='h')

        if y < num_y - 1:
            G.add_edge((x, y), (x, y + 1), weight=w, orientation='v')

    G.add_edge((8, 0), (9, 0), multi_via=2)

    # Draw mesh
    nx.draw_networkx(G, pos=pos, node_color='gray', node_size=8, edge_color='lightgray', hold=True)

    signals = {  # [terminals, ...]
        'a': [(0, 0), (8, 5), (7, 7), (6, 3)],
        'b': [(1, 1), (9, 0)],
        'c': [(3, 3), (3, 6)],
        'd': [(4, 4)],
        # [(0,9), (9,0)],
        # [(0,1), (9,2)],
        # [(1,1), (8,9), (7,4)],
        # [(1,1), (2,0)],
        # [(4,1), (5,1), (7,3)],
        # [(1,3), (4,8)],
        # [(3,3), (6,5)],
        # [(1,2), (0,0), (0,4), (2,0), (2,4), (9,9), (5,5)],
        # [(1,1), (1,8)],
        # [(3,1), (3,8)],
        # [(9,0), (0,9), (5,5)],
    }

    colors = ['red', 'blue', 'green', 'orange', 'violet']
    router = PathFinderGraphRouter(DijkstraRouter())
    routing_trees = router.route(G, signals)
    edge_labels = {(a, b): "%.2f" % data.get('weight', 0) for (a, b, data) in G.edges(data=True)}

    for (name, signal), color in zip(signals.items(), colors):
        edges = list(routing_trees[name].edges)
        nx.draw_networkx_edges(G, pos, edgelist=edges, width=4, edge_color=color)

    nx.draw_networkx_nodes(G, pos, nodelist=list(chain(*signals.values())), node_size=32, node_color='black')
    nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)

    plt.draw()
    plt.show()



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx

import operator
import klayout.db as db


class LayerOp:
    def __init__(self, op, lhs, rhs):
        self.op = op
        self.lhs = lhs
        self.rhs = rhs

    def eval(self):
        pass

    def __add__(self, other):
        return LayerOp(operator.add, self, other)

    def __or__(self, other):
        return LayerOp(operator.add, self, other)

    def __and__(self, other):
        return LayerOp(operator.iand, self, other)

    def __sub__(self, other):
        return LayerOp(operator.sub, self, other)

    def __xor__(self, other):
        return LayerOp(operator.xor, self, other)


class Leaf(LayerOp):
    def __init__(self, val):
        self.val = val

    def eval(self):
        return self.val


class AbstractLayer(Leaf):

    def __init__(self, layer_num, layer_purpose):
        self.layer_num = layer_num
        self.layer_purpose = layer_purpose

    def eval(self):
        return self.layer_num, self.layer_purpose


def layer(idx: int,
          purpose: int = 0,
          name: str = None) -> AbstractLayer:
    """ Get a handle to a layer by layer number.
    :param idx: GDS layer number or a string of the form '1/0'.
    :param purpose: GDS layer purpose.
    :param name: Name as a string.
    :return: Handle to the layer.
    """

    if name is None:
        name = '{}/{}'.format(idx, purpose)

    # Allow idx to be a string like '1/0'.
    if isinstance(idx, str):
        s = idx.split('/', 2)
        a, b = s
        idx = int(a)
        purpose = int(b)

    return AbstractLayer(idx, purpose, material=material)


class Mask:
    """ Wrapper around db.Region.
    """

    def __init__(self, region: db.Region):
        self.region = region

    def __add__(self, other):
        return Mask(self.region + other.region)

    def __or__(self, other):
        return self + other

    def __and__(self, other):
        return Mask(self.region & other.region)

    def __sub__(self, other):
        m = Mask(self.region - other.region)
        m.material = self.material
        return m

    def __xor__(self, other):
        return Mask(self.region ^ other.region)

    def __hash__(self):
        return hash(self.region)

    def __equal__(self, other):
        return self.region == other.region


def eval_op_tree(cell: db.Cell, op_node: LayerOp) -> Mask:
    """ Recursively evaluate the layer operation tree.
    :param op_node: Operand node or leaf.
    :return: Returns a `Mask` object containing a `pya.Region` of the layer.
    """

    if isinstance(op_node, AbstractLayer):
        (idx, purpose) = op_node.eval()
        layer_index = layout.layer(idx, purpose)

        region = _flatten_cell(cell, layer_index, selection_box=selection_box)
        result = Mask(region)
    else:
        assert isinstance(op_node, LayerOp)
        op = op_node.op
        lhs = eval_op_tree(cell, op_node.lhs)
        rhs = eval_op_tree(cell, op_node.rhs)
        result = op(lhs, rhs)
    result.region.merge()

    return result


l_ndiffusion = 'ndiffusion'
l_pdiffusion = 'pdiffusion'
l_nwell = 'nwell'
l_pwell = 'pwell'
l_poly = 'poly'
l_poly_label = 'poly_label'
l_diff_contact = 'diff_contact'
l_poly_contact = 'poly_contact'
l_metal1 = 'metal1'
l_metal1_label = 'metal1_label'
l_metal1_pin = 'metal1_pin'
l_via1 = 'via1'
l_metal2 = 'metal2'
l_metal2_label = 'metal2_label'
l_metal2_pin = 'metal2_pin'
l_abutment_box = 'abutment_box'

layermap = {
    l_nwell: (1, 0),
    l_pwell: (2, 0),
    l_ndiffusion: (3, 0),
    l_pdiffusion: (4, 0),
    l_poly: (5, 0),
    l_diff_contact: (6, 0),
    l_poly_contact: (7, 0),
    l_metal1: (8, 0),
    l_metal1_label: (8, 1),
    l_metal1_pin: (8, 2),
    l_via1: (9, 0),
    l_metal2: (10, 0),
    l_metal2_label: (10, 1),
    l_metal2_pin: (10, 2),
    l_abutment_box: (100, 0)
}

layermap_reverse = {v: k for k, v in layermap.items()}

via_layers = nx.Graph()
via_layers.add_edge(l_ndiffusion, l_metal1, layer=l_diff_contact)
via_layers.add_edge(l_pdiffusion, l_metal1, layer=l_diff_contact)
via_layers.add_edge(l_poly, l_metal1, layer=l_poly_contact)
via_layers.add_edge(l_metal1, l_metal2, layer=l_via1)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx
from itertools import chain, combinations, product

from typing import Any, Dict, List, Set, Tuple, AbstractSet, Optional

import logging
from .graphrouter import GraphRouter

logger = logging.getLogger(__name__)


class HVGraphRouter(GraphRouter):

    def __init__(self, sub_graphrouter: GraphRouter, orientation_change_penalty: float = 1):
        self.sub_graphrouter = sub_graphrouter
        self.orientation_change_penalty = orientation_change_penalty

    def route(self,
              graph: nx.Graph,
              signals: Dict[Any, List[Any]],
              reserved_nodes: Optional[Dict[Any, AbstractSet[Any]]] = None,
              node_conflict: Optional[Dict[Any, AbstractSet[Any]]] = None,
              equivalent_nodes: Optional[Dict[Any, AbstractSet[Any]]] = None,
              is_virtual_node_fn=None
              ) -> Dict[Any, nx.Graph]:
        return _route_hv(self.sub_graphrouter,
                         graph,
                         signals=signals,
                         reserved_nodes=reserved_nodes,
                         node_conflict=node_conflict,
                         is_virtual_node_fn=is_virtual_node_fn)


def _build_hv_routing_graph(graph: nx.Graph, orientation_change_penalty=1) -> Tuple[nx.Graph, Dict, Dict]:
    """ Apply a transformation to G to simplify corner avoidance routing.
    Essentially nodes with edges of different orientations are split into multiple virtual nodes,
    one per edge orientation.
    The virtual nodes are connected by virtual edges of weight `orientation_change_penalty`.

    Use `flatten_hv_graph` to transform a graph/subgraph back to its original form.

    Parameters
    ----------
    graph: Routing graph with edge orientation information.
            Edge orientation must be stored in the 'orientation' field of the networkx edge data.
    orientation_change_penalty: Cost for changes between different orientations.

    :return : Tuple[nx.Graph, Dict[node, Dict[orientation, node]], Dict[node, node]]
    """

    assert nx.is_connected(graph), Exception("Graph G is not connected.")

    H = nx.Graph()

    # G -> H
    node_mapping = dict()

    # Create new nodes.
    for n1 in graph:
        edges = graph.edges(n1, data=True)
        orientations = set(data.get('orientation', None) for a, b, data in edges)
        # orientations = {o for o in orientations if o is not None}
        orientations.add(None)
        has_different_orientations = len(orientations) > 1

        # Add edges between virtual nodes.
        if has_different_orientations:
            # Create a virtual node for all orientations.
            n2s = {o: (o, n1) for o in orientations}
            node_mapping[n1] = n2s
            assert (len(n2s.values()) >= 2)
            # Mutually connect virtual nodes.
            for a, b in combinations(n2s.values(), 2):
                w = orientation_change_penalty
                if a is None or b is None:
                    w = 0
                H.add_edge(a, b, weight=w)
        else:
            # No need to split node.
            n2 = (None, n1)
            n2s = {None: n2}
            node_mapping[n1] = n2s
            H.add_node(n2)

    # Create new edges.
    for a, b, data in graph.edges(data=True):

        orientation = data.get('orientation', None)
        all_a2 = node_mapping[a]
        all_b2 = node_mapping[b]

        if orientation in all_a2 and orientation in all_b2:
            a2 = all_a2[orientation]
            b2 = all_b2[orientation]
            H.add_edge(a2, b2, **data)
        else:
            for a2, b2 in product(all_a2.values(), all_b2.values()):
                H.add_edge(a2, b2, **data)

    for n in H.nodes:
        assert nx.degree(H, n) > 0, Exception("Unconnected node %s" % str(n))

    assert nx.is_connected(H), Exception("Graph is not connected.")

    # Create reverse mapping.
    reverse_mapping = dict()
    for n1, n2s in node_mapping.items():
        for n2 in n2s.values():
            reverse_mapping[n2] = n1

    return H, node_mapping, reverse_mapping


def _flatten_hv_graph(hv_graph: nx.Graph, reverse_mapping: Dict) -> nx.Graph:
    """ Inverse transformation of `build_hv_routing_graph`.
    """

    G = nx.Graph()

    for n in hv_graph:
        n2 = reverse_mapping[n]
        G.add_node(n2)

    for a, b, data in hv_graph.edges(data=True):
        a2 = reverse_mapping[a]
        b2 = reverse_mapping[b]
        if a2 != b2:
            G.add_edge(a2, b2, **data)

    return G


def _route_hv(router: GraphRouter,
              graph: nx.Graph,
              signals: Dict[Any, List[Any]],
              orientation_change_penalty: float = 1,
              node_conflict: Dict[Any, Set[Any]] = None,
              reserved_nodes: Optional[Dict[Any, AbstractSet[Any]]] = None,
              is_virtual_node_fn=None,
              **kw) -> Dict[Any, nx.Graph]:
    """ Global routing with corner avoidance.
    Corners (changes between horizontal/vertical tracks) are avoided by transforming the routing graph `G`
    such that corners are penalized.

    Parameters
    ----------
    :param graph: Routing graph with edge orientation information.
            Edge orientation must be stored in the 'orientation' field of the networkx edge data.
    :param signals: A dict mapping signal names to signal terminals.
    :param orientation_change_penalty: Cost for changes between different orientations.
    :param reserved_nodes: An optional dict which specifies nodes that are reserved for a specific net.
    Dict[net_name, set of nodes].
    :param kw: Parameters to be passed to underlying routing function.

    """

    assert isinstance(signals, dict)
    logger.debug('Start global routing with corner avoidance.')

    H, node_mapping, node_mapping_reverse = _build_hv_routing_graph(
        graph,
        orientation_change_penalty=orientation_change_penalty
    )
    reserved_nodes_h = None
    if reserved_nodes is not None:
        reserved_nodes_h = {net: list(chain(*(node_mapping[n].values() for n in nodes))) for net, nodes in
                            reserved_nodes.items()}

    if node_conflict is None:
        node_conflict = dict()

    # For each node find other nodes that are equivalent when mapped back.
    equivalent_nodes = {
        n_h: set(node_mapping[n_g].values()) for n_h, n_g in node_mapping_reverse.items()
    }

    # Some nodes in H will be mapped to the same node in G and therefore conflict with each other.
    node_conflict_h = dict()

    for n_h in H:
        n_g = node_mapping_reverse[n_h]

        conflicts = set()
        conflicts.update(node_mapping[n_g].values())

        if n_g in node_conflict:
            conflicts_g = set(node_conflict[n_g])
            for n in conflicts_g:
                conflicts.update(node_mapping[n].values())

        node_conflict_h[n_h] = conflicts

    signals_h = {net: [node_mapping[t][None] for t in terminals] for net, terminals in signals.items()}

    def _is_virtual_node_fn(n) -> bool:
        return is_virtual_node_fn(node_mapping_reverse[n])

    assert nx.is_connected(H)
    routing_trees_h = router.route(H, signals_h,
                                   reserved_nodes=reserved_nodes_h,
                                   node_conflict=node_conflict_h,
                                   equivalent_nodes=equivalent_nodes,
                                   is_virtual_node_fn=_is_virtual_node_fn,
                                   **kw)

    # logger.info("Use pyo3 router.")
    # from . import pyo3_graphrouter
    # routing_trees_h = pyo3_graphrouter.route(H, signals_h, node_conflict=node_conflict_h)

    # Translate routing trees from H back to G.
    routing_trees = {name: _flatten_hv_graph(rt, node_mapping_reverse)
                     for name, rt in routing_trees_h.items()}

    # Assert that reserved_nodes is respected.
    # A node that is reserved for a signal should not be used by another signal.
    if reserved_nodes:
        for net, nodes in reserved_nodes.items():
            for rt_net, rt in routing_trees.items():
                if net != rt_net:
                    for n in nodes:
                        assert n not in rt.nodes, \
                            "Node %s is reserved for net %s but has been used for net %s." % (
                                n, net, rt_net)

    return routing_trees



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from z3 import *
from .place import TransistorPlacer
from typing import Iterable
from ..data_types import Transistor, Cell, ChannelType
from itertools import combinations, chain
import logging

logger = logging.getLogger(__name__)


class SMTPlacer(TransistorPlacer):

    def __init__(self):
        pass

    def place(self, transistors: Iterable[Transistor]) -> Cell:
        """
        Place transistors using an SMT solver (Z3).
        :param transistors:
        :return: Placed cell.
        """
        transistors = list(transistors)
        nmos = [t for t in transistors if t.channel_type == ChannelType.NMOS]
        pmos = [t for t in transistors if t.channel_type == ChannelType.PMOS]

        # TODO: expose this as a parameter for multi row cells.
        max_cell_rows = 1

        # Upper bound on cell width.
        max_cell_width = (max(len(nmos), len(pmos)) * 2) // max_cell_rows

        # Optimizer
        opt = Optimize()

        # Wrapper arount opt.add
        def add_assertion(assertion, **kw):
            opt.add(assertion)

        # Create symbols for bounds on transistor positions.
        # Used later to minimize cell width.
        max_x = Int("max_x")
        max_y = Int("max_y")

        # Create symbols for transistor positions.
        transistor_positions = {t: (Int("transistor_{}_x".format(i)), Int("transistor_{}_y".format(i)))
                                for i, t in enumerate(transistors)}

        # Create boolean symbols for transistor flips.
        # Each transistor can be flipped (source/drain swapped).
        transistor_flipped = {t: Bool("transistor_{}_flipped".format(i))
                              for i, t in enumerate(transistors)}

        # Constraint: Positions are bounded.
        # Add bounds on positions.
        for x, y in transistor_positions.values():
            add_assertion(x >= 0)
            add_assertion(y >= 0)

            # Add upper bounds on transistor positions.
            add_assertion(x < max_cell_width)
            add_assertion(y < max_cell_rows * 2)

            add_assertion(max_x >= x)
            add_assertion(max_y >= y)

        # Constraint: Separate rows for NMOS and PMOS
        # Assign rows to NMOS and PMOS
        for t, (x, y) in transistor_positions.items():

            or_constraints = []

            for r in range(max_cell_rows):
                # Place transistor in upper or lower stack?
                # Ordering alternates from row to row such that power stripe can be shared.
                stack = r % 2 if t.channel_type == ChannelType.NMOS else 1 - r % 2

                allowed_y = r * 2 + stack
                or_constraints.append(y == allowed_y)

            on_allowed_row = Or(*or_constraints)
            add_assertion(on_allowed_row)

        # Constraint: Non-overlapping positions
        # No two transistors should have the same position.
        for (x1, y1), (x2, y2) in combinations(transistor_positions.values(), 2):
            same_position = And(
                x1 == x2,
                y1 == y2
            )
            different_positions = Not(same_position)
            add_assertion(different_positions)

        # Constraint: Diffusion sharing
        # If two transistors are placed side-by-side then the abutted sources/drain nets must match.
        for ts in [nmos, pmos]:
            # Loop through all potential (left, right) pairs.
            for a, b in combinations(ts, 2):
                for t_left, t_right in [(a, b), (b, a)]:
                    xl, yl = transistor_positions[t_left]
                    xr, yr = transistor_positions[t_right]

                    # Checks if t_left is left neighbor of t_right.
                    are_neighbors = And(
                        yl == yr,
                        xl + 1 == xr
                    )

                    # Go through all combinations of flipped transistors
                    # and check if they are allowed to be directly abutted if flipped
                    # in a specific way.
                    flip_combinations = [[False, False], [False, True], [True, False], [True, True]]
                    for flip_l, flip_r in flip_combinations:
                        l = t_left.flipped() if flip_l else t_left
                        r = t_right.flipped() if flip_r else t_right

                        if l.right != r.left:
                            # Drain/Source net mismatch.
                            # In case the transistors are flipped that way,
                            # they are not allowed to be direct neighbors.
                            add_assertion(
                                Implies(
                                    And(transistor_flipped[t_left] == flip_l,
                                        transistor_flipped[t_right] == flip_r),
                                    Not(are_neighbors)
                                )
                            )

        # Extract all net names.
        nets = set(chain(*(t.terminals() for t in transistors)))

        # Create net bounds. This will be used to optimize
        # the bounding box perimeter of the nets (for wiring length optimization).
        net_max_x = {net: Int("net_max_x_{}".format(net))
                     for net in nets}

        net_min_x = {net: Int("net_min_x_{}".format(net))
                     for net in nets}

        net_max_y = {net: Int("net_max_y_{}".format(net))
                     for net in nets}

        net_min_y = {net: Int("net_min_y_{}".format(net))
                     for net in nets}

        for t in transistors:
            x, y = transistor_positions[t]

            # TODO: Net positions dependent on transistor terminal.
            #       Now, the net position equals the transistor position.
            #       Make it dependent on the actual terminal (drain, gate, source).
            #       Also depends on transistor flips.
            for net in t.terminals():
                add_assertion(x <= net_max_x[net])
                add_assertion(x >= net_min_x[net])
                add_assertion(y <= net_max_y[net])
                add_assertion(y >= net_min_y[net])

        # Optiimization goals
        # Note: z3 uses lexicographic priorities of objectives by default.
        # Here, the cell width is optimized first.
        # Could be interesting: z3 could also find pareto fronts.

        # Optimization objective 1
        # Minimize cell width.
        opt.minimize(max_x)

        # Optimization objective 2
        # Minimize wiring length (net bounding boxes)
        # TODO: sort criteria by what? Number of terminals?
        for net in nets:
            # TODO: skip VDD/GND nets
            opt.minimize(net_max_x[net] - net_min_x[net])
            opt.minimize(net_max_y[net] - net_min_y[net])

        # TODO: optimization objective for pin nets.

        logger.info("Run SMT optimizer (Z3)")
        sat = opt.check()

        logger.info("SMT result: %s", "Satisfied" if sat else "Unsatisfied")
        if not sat:
            msg = "Placement problem not satisfiable."
            logger.error(msg)
            raise Exception(msg)

        # logger.debug("model = %s", opt.model())

        model = opt.model()
        cell_width = model[max_x].as_long() + 1

        cell = Cell(cell_width)
        rows = [cell.lower, cell.upper]
        for t in transistors:
            x, y = transistor_positions[t]
            x = model[x].as_long()
            y = model[y].as_long()
            flip = is_true(model[transistor_flipped[t]])

            flipped = t.flipped() if flip else t

            rows[y][x] = flipped

        return cell


def test():
    placer = SMTPlacer()
    from itertools import count
    c = count()
    transistors = [Transistor(ChannelType.PMOS, 1, 1, 3, name=next(c)),
                   Transistor(ChannelType.NMOS, 1, 2, 3, name=next(c)),
                   Transistor(ChannelType.PMOS, 1, 1, 3, name=next(c)),
                   Transistor(ChannelType.NMOS, 1, 2, 3, name=next(c)),
                   Transistor(ChannelType.PMOS, 1, 1, 3, name=next(c)),
                   Transistor(ChannelType.NMOS, 1, 2, 3, name=next(c))]
    placer.place(transistors)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import logging
import time
from typing import Dict, List, Tuple, Union
from klayout import db
import os

from .writer import Writer, remap_layers
from ..layout import layers

logger = logging.getLogger(__name__)


class GdsWriter(Writer):
    """
    Writer for GDS2 output.
    """

    def __init__(self,
                 db_unit: float,
                 output_map: Dict[str, Tuple[int, int]]):
        self.db_unit = db_unit
        self.output_map = output_map

    def write_layout(self,
                     layout: db.Layout,
                     pin_geometries: Dict[str, List[Tuple[str, db.Shape]]],
                     top_cell: db.Cell,
                     output_dir: str,
                     ) -> None:
        # Re-map layers
        layout = remap_layers(layout, self.output_map)

        # Set database unit.
        # klayout expects dbu to be in µm, the tech file takes it in meters.
        layout.dbu = self.db_unit * 1e6
        logger.debug("dbu = {} µm".format(layout.dbu))

        # Possibly scale the layout.
        scaling_factor = 1
        if scaling_factor != 1:
            logger.info("Scaling layout by factor {}".format(scaling_factor))
            layout.transform(db.DCplxTrans(scaling_factor))

        # Write GDS.
        gds_file_name = '{}.gds'.format(top_cell.name)
        gds_out_path = os.path.join(output_dir, gds_file_name)
        logger.info("Write GDS: %s", gds_out_path)
        layout.write(gds_out_path)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from enum import Enum

from itertools import islice, tee, chain, product
from copy import deepcopy
from typing import Any, Set, Tuple


class ChannelType(Enum):
    NMOS = 1,
    PMOS = 2


class Transistor:
    """
    Abstract representation of a MOS transistor.
    """

    def __init__(self, channel_type: ChannelType,
                 source_net: str, gate_net: str, drain_net: str,
                 channel_width=None,
                 name: str = 'M?',
                 allow_flip_source_drain: bool = True
                 ):
        """
        params:
        left: Either source or drain net.
        right: Either source or drain net.
        """
        self.name = name
        self.channel_type = channel_type
        self.source_net = source_net
        self.gate_net = gate_net
        self.drain_net = drain_net

        self.channel_width = channel_width

        self.allow_flip_source_drain = allow_flip_source_drain

        # TODO
        self.threshold_voltage = None

    def flipped(self):
        """ Return the same transistor but with left/right terminals flipped.
        """

        assert self.allow_flip_source_drain, "Flipping source and drain is not allowed."

        f = deepcopy(self)
        f.source_net = self.drain_net
        f.drain_net = self.source_net

        return f

    def terminals(self) -> Tuple[Any, Any, Any]:
        """ Return a tuple of all terminal names.
        :return:
        """
        return self.source_net, self.gate_net, self.drain_net

    def __key(self):
        return self.name, self.channel_type, self.source_net, self.gate_net, self.drain_net, self.channel_width, self.threshold_voltage

    def __hash__(self):
        return hash(self.__key())

    def __eq__(x, y):
        return x.__key() == y.__key()

    def __repr__(self):
        return "({}, {}, {})".format(self.source_net, self.gate_net, self.drain_net)


class Cell:
    """ Dual row cell.
    """

    def __init__(self, width: int):
        self.width = width
        self.upper = [None] * width
        self.lower = [None] * width

    def get_transistor_locations(self) -> Set[Tuple[Transistor, Tuple[int, int]]]:
        """ Get a list of all transistors together with their location.
        Transistor locations are given on a grid like:

         | (0,1) | (1,1) | ...
         | (0,0) | (1,0) | ...

        Returns
        -------

        Returns a set of (transistor, (x,y)).
        """

        assert len(self.lower) == len(self.upper)

        t = [self.lower, self.upper]
        idx = product(range(self.width), range(2))

        return set((t[y][x], (x, y)) for x, y in idx if t[y][x] is not None)

    def __repr__(self):
        """ Pretty-print
        """

        return (
                " | ".join(['{:^16}'.format(str(t)) for t in self.upper]) +
                "\n" +
                " | ".join(['{:^16}'.format(str(t)) for t in self.lower])
        )



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx
from .signal_router import SignalRouter

from typing import Any, Dict, List, Optional, Union, AbstractSet
from itertools import product, chain, count
import logging

# Import GLPK binding
from pulp import *
from pulp.constants import LpStatus

logger = logging.getLogger(__name__)


class LPSignalRouter(SignalRouter):

    def __init__(self):
        pass

    def route(self, G: nx.Graph,
              terminals: List[Any],
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:
        return min_steiner_trees(
            G,
            {0: terminals},
            node_cost_fn,
            edge_cost_fn,
        )[0]


class LPGraphRouter():

    def route(self, G: nx.Graph,
              signals: Dict[Any, List[Any]],
              reserved_nodes: Optional[Dict] = None,
              node_conflict: Optional[Dict[Any, AbstractSet[Any]]] = None,
              node_cost_fn=None,
              edge_cost_fn=None
              ) -> Dict[Any, nx.Graph]:
        assert isinstance(signals, dict)

        if edge_cost_fn is None:
            edge_cost_fn = lambda x: G[x[0]][x[1]]['weight']

        routing_trees = min_steiner_trees(
            G,
            signals,
            node_cost_fn=node_cost_fn,
            edge_cost_fn=edge_cost_fn,
            reserved_nodes=reserved_nodes,
            node_conflict=node_conflict
        )

        return routing_trees


def min_steiner_trees(
        G: nx.Graph,
        signals: Dict[Any, AbstractSet[Any]],
        node_cost_fn,
        edge_cost_fn,
        reserved_nodes: Optional[Dict[Any, AbstractSet[Any]]] = None,
        node_conflict: Optional[Dict[Any, AbstractSet[Any]]] = None
) -> Dict[Any, nx.Graph]:
    """
    Find a minimum weight Steiner trees which connects all `terminals` of all `signals` in `G`
    but without creating shorts between the signals.

    :param G: Graph.
    :param terminals: List of terminals to be connected.

    References:
    Sunil Chopra and Chih-Yang Tsai, Polyhedral approaches for the Steiner Tree on graphs, 2000.
    """

    if node_conflict is None:
        node_conflict = dict()

    if reserved_nodes is None:
        reserved_nodes = dict()

    # Get solver instance.
    problem = LpProblem("steiner_tree", LpMinimize)

    def add_assertion(assertion: LpConstraint, problem=problem):
        """
        Add assertion to solver.
        :param assertion:
        :param kwargs:
        :return:
        """
        problem += assertion

    _counter = count()

    def fresh_variable(template: str = 'var%d') -> LpVariable:
        """
        Create fresh integer variable.
        :param template:
        :return:
        """
        return LpVariable(template % next(_counter), cat='Continuous')

    def fresh_int_variable(template: str = 'var%d') -> LpVariable:
        """
        Create fresh integer variable.
        :param template:
        :return:
        """
        return LpVariable(template % next(_counter), cat='Integer')

    # def create_constraints_multi_commodity_flow(problem: LpProblem, source, sinks) -> Dict[Any, LpVariable]:
    #     """
    #     Create constraints for steiner tree of a single signal.
    #
    #     Experiment: model a flow from one source to multiple sinks. This has the advantage of using less
    #     variables and less constraints but shows to be slower in practice that modelling multiple
    #     single-commodity flows.
    #
    #     :param source:
    #     :param sinks:
    #     :return:
    #     """
    #     # Edge choice variables.
    #     x = {(i, j): fresh_int_variable(template='x%d')
    #          for i, j in G.edges()
    #          }
    # 
    #     # Flow variables
    #     f = dict()
    #     f.update({(i, j): fresh_variable(template='f%d')
    #               for (i, j) in G.edges
    #               })
    #     f.update({(j, i): fresh_variable(template='f%d')
    #               for (i, j) in G.edges
    #               })
    #
    #     y = dict()
    #     y.update({(i, j): fresh_variable(template='y%d') for i, j in G.edges})
    #     y.update({(j, i): fresh_variable(template='y%d') for i, j in G.edges})
    #
    #     # Constraint (1)
    #     for j in G.nodes:
    #         left = sum((f[(i, j)] - f[(j, i)] for i in G.neighbors(j)))
    #
    #         if j == source:
    #             right = -1 * len(sinks)
    #         elif j in sinks:
    #             right = 1
    #         else:
    #             right = 0
    #
    #         add_assertion(left == right)
    #
    #     # Constraint (3)
    #     for (i, j) in G.edges:
    #         scale = 1 / len(sinks)
    #         add_assertion(f[(i, j)] * scale <= y[(i, j)])
    #         add_assertion(f[(j, i)] * scale <= y[(j, i)])
    #
    #     # Constraint (4)
    #     for i, j in G.edges:
    #         y_ij = y[(i, j)]
    #         y_ji = y[(j, i)]
    #         x_ij = x[(i, j)]
    #         add_assertion(
    #             y_ij + y_ji == x_ij
    #         )
    #
    #     # f_kij must be >= 0
    #     for f_ij in f.values():
    #         add_assertion(f_ij >= 0)
    #
    #     return x

    def create_constraints_single_commodity_flow(problem: LpProblem, source, sinks) -> Dict[Any, LpVariable]:
        """
        Create constraints for steiner tree of a single signal.
        :param source:
        :param sinks:
        :return:
        """
        # Edge choice variables.
        x = {(i, j): fresh_int_variable(template='x%d')
             for i, j in G.edges()
             }

        # Flow variables
        f = dict()
        f.update({(k, i, j): fresh_variable(template='f%d')
                  for k, (i, j) in product(sinks, G.edges)
                  })
        f.update({(k, j, i): fresh_variable(template='f%d')
                  for k, (i, j) in product(sinks, G.edges)
                  })

        y = dict()
        y.update({(i, j): fresh_variable(template='y%d') for i, j in G.edges})
        y.update({(j, i): fresh_variable(template='y%d') for i, j in G.edges})

        # Constraint (1)
        # For each source/sink pair there must be a flow of value 1 from source to sink k.
        # The net flow for non-sink and non-source nodes must be 0.
        # The net flow for the source node must be -1 and for the sink node +1.
        for k in sinks:
            for j in G.nodes:
                left = sum((f[(k, i, j)] - f[(k, j, i)] for i in G.neighbors(j)))

                if j == source:
                    right = -1
                elif j == k:
                    right = 1
                else:
                    right = 0

                add_assertion(left == right)

        # Constraint (3)
        for k, (i, j) in product(sinks, G.edges):
            add_assertion(f[(k, i, j)] <= y[(i, j)])
            add_assertion(f[(k, j, i)] <= y[(j, i)])

        # Constraint (4)
        for i, j in G.edges:
            y_ij = y[(i, j)]
            y_ji = y[(j, i)]
            x_ij = x[(i, j)]
            add_assertion(
                y_ij + y_ji == x_ij
            )

        # f_kij must be >= 0
        for f_kij in f.values():
            add_assertion(f_kij >= 0)

        return x

    def create_node_usage_indicator(edge_usage: Dict[Any, LpVariable]) -> Dict[Any, LpVariable]:
        """
        Create variables indicating if a node is used based on usage of edges.
        :param edge_usage:
        :return:
        """
        # Introduce variable for indicating node usage.
        node_used = {n: fresh_variable(template='node_used_%d')
                     for n in G.nodes}
        # node_used can be either 0 or 1
        for n, var in node_used.items():
            add_assertion(var >= 0)
            add_assertion(var <= 1)
        # node_used must be 1 if an adjacent edge is used.
        for i, j in G.edges:
            add_assertion(
                node_used[i] >= edge_usage[(i, j)]
            )
            add_assertion(
                node_used[j] >= edge_usage[(i, j)]
            )

        return node_used

    total_cost = 0

    used_edges_by_signal = dict()
    # `node_sharing` holds for each node the number of signals that are using the node.
    if len(signals) > 1:
        node_sharing = {n: 0 for n in G.nodes}
    else:
        node_sharing = dict()

    used_nodes_by_signal = dict()

    # For each signal:
    # * create connectivity constraint
    for signal_name, terminals in signals.items():
        source = terminals[0]
        sinks = terminals[1:]
        used_edges = create_constraints_single_commodity_flow(problem, source, sinks)
        # used_edges = create_constraints_multi_commodity_flow(problem, source, sinks)
        used_edges_by_signal[signal_name] = used_edges

        nodes_used = create_node_usage_indicator(used_edges)

        # Add constraint for node conflicts between different signals.
        # If node n1 is used, then mark all conflicting nodes also as used.
        for n1, conflicts in node_conflict.items():
            for n2 in conflicts:
                problem += nodes_used[n2] >= nodes_used[n1]
                problem += nodes_used[n1] >= nodes_used[n2]

        used_nodes_by_signal[signal_name] = nodes_used

        # Count by how many signals a node is used.
        if len(signals) > 1:
            for n, var in nodes_used.items():
                node_sharing[n] += var

        if node_cost_fn is not None:
            # Node weights.

            node_weight = sum(
                node_cost_fn(n) * nodes_used[n]
                for n in G.nodes
            )

            total_cost += node_weight

        if edge_cost_fn is not None:
            edge_weight = sum(
                edge_cost_fn((i, j)) * used_edges[(i, j)]
                for i, j, data in G.edges(data=True)
            )

            total_cost += edge_weight

    # Distinct nodes constraint.
    # Assert that a node cannot be shared among signals.
    for n, num_signals in node_sharing.items():
        add_assertion(num_signals <= 1)

    # Add constraints for reserved nodes.
    all_nets = set(signals.keys())
    for net_name, reserved in reserved_nodes.items():
        other_nets = all_nets - {net_name}
        # If a node is reserved for some signal, then the other signals can't use it.
        for net in other_nets:
            for r in reserved:
                problem += used_nodes_by_signal[net][r] == 0

    # Minimize sum of edge weights.
    problem.setObjective(total_cost)

    logger.info('Invoke LP solver')
    problem.solve()
    logger.info('LP solver result: %s', LpStatus[problem.status])

    if problem.status == 1:
        # Map solution back to a routing graph.
        routing_trees = dict()
        for net_name, used_edges in used_edges_by_signal.items():
            routing_tree = nx.Graph()
            for i, j in G.edges:
                x_ij = used_edges[(i, j)]
                value = x_ij.varValue
                if value > 0:
                    routing_tree.add_edge(i, j)
            routing_trees[net_name] = routing_tree

        return routing_trees

    else:
        msg = 'Problem not solved: {}'.format(LpStatus[problem.status])
        logger.error(msg)
        raise Exception(msg)


def test_lp_route():
    logging.basicConfig(level=logging.INFO)
    import matplotlib.pyplot as plt

    G = nx.Graph()

    num_x = 10
    num_y = 10
    x = range(0, num_x)
    y = range(0, num_y)

    # Store drawing positions of vertices. For plotting only.
    pos = {}

    # Construct mesh
    for i, (x, y) in enumerate(product(x, y)):
        G.add_node((x, y))
        pos[(x, y)] = (x, y)

        if x < num_x - 1 and not (1 <= y < 5 and x == 4):
            G.add_edge((x, y), (x + 1, y), weight=1, orientation='h')

        if y < num_y - 1:
            G.add_edge((x, y), (x, y + 1), weight=1, orientation='v')

    # Draw mesh
    nx.draw_networkx(G, pos=pos, node_color='gray', node_size=8, edge_color='lightgray', hold=True)

    signals = {  # [terminals, ...]
        'a': [(0, 0), (8, 5), (7, 7), (6, 3)],
        'b': [(1, 1), (9, 0)],
        # [(3, 3), (3, 6)],
        # [(0, 9), (9, 0)],
        # [(0,1), (9,2)],
        # [(1,1), (8,9), (7,4)],
        # [(1,1), (2,0)],
        # [(4,1), (5,1), (7,3)],
        # [(1,3), (4,8)],
        # [(3,3), (6,5)],
        # [(1,2), (0,0), (0,4), (2,0), (2,4), (9,9), (5,5)],
        # [(1,1), (1,8)],
        # [(3,1), (3,8)],
        # [(9,0), (0,9), (5,5)],
    }

    conflicts = {(1, 0): {(0, 1)}}

    reserved_nodes = {
        'a': {(3, 0), (0, 1)}
    }

    routing_trees = min_steiner_trees(G, signals,
                                      node_cost_fn=None,
                                      edge_cost_fn=lambda e: 1,
                                      node_conflict=conflicts,
                                      reserved_nodes=reserved_nodes)

    colors = ['red', 'blue', 'green', 'orange', 'violet']
    # routing_trees = route_hv(DijkstraRouter(), G, signals, orientation_change_penalty=10)

    # edge_labels = {(a, b): "%.2f" % data.get('weight', 0) for (a, b, data) in G.edges(data=True)}

    # nx.draw_networkx_edges(G, pos, edgelist=routing_tree.edges, width=4, edge_color=colors[0])

    for i, (signal_name, signal) in enumerate(signals.items()):
        edges = list(routing_trees[signal_name].edges)
        nx.draw_networkx_edges(G, pos, edgelist=edges, width=4, edge_color=colors[i])

    nx.draw_networkx_nodes(G, pos, nodelist=list(chain(*signals.values())), node_size=32, node_color='black')
    # nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)

    plt.draw()
    plt.show()



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from itertools import chain
from collections import Counter, defaultdict
import numpy
import toml

from lccommon import net_util
from lccommon.net_util import load_transistor_netlist, is_ground_net, is_supply_net
from lclayout.data_types import Cell

from .place.place import TransistorPlacer
from .place.euler_placer import EulerPlacer, HierarchicalPlacer
from .place.smt_placer import SMTPlacer

from .graphrouter.graphrouter import GraphRouter
from .graphrouter.hv_router import HVGraphRouter
from .graphrouter.pathfinder import PathFinderGraphRouter
from .graphrouter.signal_router import DijkstraRouter, ApproxSteinerTreeRouter

from .layout.transistor import *
from .layout import cell_template
from .layout.notch_removal import fill_notches

from .routing_graph import *

from .drc_cleaner import drc_cleaner
from .lvs import lvs

# klayout.db should not be imported if script is run from KLayout GUI.
if 'pya' not in sys.modules:
    import klayout.db as pya

logger = logging.getLogger(__name__)


def _merge_all_layers(shapes):
    # TODO: Move into LcLayout class.
    # Merge all polygons on all layers.
    for layer_name, s in shapes.items():
        if '_label' not in layer_name:
            r = pya.Region(s)
            r.merge()
            s.clear()
            s.insert(r)


def _draw_label(shapes, layer, pos: Tuple[int, int], text: str) -> None:
    """
    Insert a pya.Text object into `shapes`.
    :param shapes:
    :param layer:
    :param pos: Position of the text as a (x,y)-tuple.
    :param text: Text.
    :return: None
    """
    x, y = pos
    # shapes[layer].insert(pya.Text.new(text, pya.Trans(x, y), 0.1, 2))
    shapes[layer].insert(pya.Text.new(text, x, y))


def _draw_routing_tree(shapes: Dict[str, pya.Shapes],
                       G: nx.Graph,
                       rt: nx.Graph,
                       tech,
                       debug_routing_graph: bool = False):
    """ Draw a routing graph into a layout.
    :param shapes: Mapping from layer name to pya.Shapes object
    :param G: Full graph of routing grid
    :param rt: Graph representing the wires
    :param tech: module containing technology information
    :param debug_routing_graph: Draw narrower wires for easier visual inspection
    :return:
    """

    def is_virtual_node(n):
        return n[0].startswith('virtual')

    def is_virtual_edge(e):
        return is_virtual_node(e[0]) or is_virtual_node(e[1])

    logger.debug("Drawing wires")

    # Loop through all edges of the routing tree and draw them individually.
    for a, b in rt.edges:

        if not is_virtual_edge((a, b)):

            l1, (x1, y1) = a
            l2, (x2, y2) = b

            data = G[a][b]

            if l1 == l2:
                # On the same layer -> wire

                w = data.get('wire_width', tech.wire_width[l1])

                ext = w // 2

                is_horizontal = y1 == y2 and x1 != x2

                if is_horizontal:
                    w = tech.wire_width_horizontal[l1]

                if debug_routing_graph:
                    w = min(tech.routing_grid_pitch_x, tech.routing_grid_pitch_y) // 16

                path = pya.Path([pya.Point(x1, y1), pya.Point(x2, y2)], w, ext, ext)
                shapes[l1].insert(path)
            else:
                # l1 != l1 -> this looks like a via
                assert x1 == x2
                assert y1 == y2
                # Draw via
                via_layer = via_layers[l1][l2]['layer']
                logger.debug('Draw via: {} ({}, {})'.format(via_layer, x1, y1))

                via_width = tech.via_size[via_layer]

                if debug_routing_graph:
                    via_width = min(tech.routing_grid_pitch_x, tech.routing_grid_pitch_y) // 16

                w = via_width // 2
                via = pya.Box(pya.Point(x1 - w, y1 - w),
                              pya.Point(x1 + w, y1 + w))
                shapes[via_layer].insert(via)

                # Ensure minimum via enclosure.
                if not debug_routing_graph:
                    for l in (l1, l2):
                        # TODO: Check on which sides minimum enclosure is not yet satisfied by some wire.

                        neighbors = rt.neighbors((l, (x1, y1)))
                        neighbors = [n for n in neighbors if n[0] == l]

                        w_ext = via_width // 2 + tech.minimum_enclosure[(l, via_layer)]
                        w_noext = via_width // 2

                        # Check on which sides the enclosure must be extended.
                        # Some sides will already be covered by a routing wire.
                        ext_right = w_ext
                        ext_upper = w_ext
                        ext_left = w_ext
                        ext_lower = w_ext
                        # TODO
                        # for _, (n_x, n_y) in neighbors:
                        #     if n_x == x1:
                        #         if n_y < y1:
                        #             ext_lower = w_noext
                        #         if n_y > y1:
                        #             ext_upper = w_noext
                        #     if n_y == y1:
                        #         if n_x < x1:
                        #             ext_left = w_noext
                        #         if n_x > x1:
                        #             ext_right = w_noext

                        enc = pya.Box(
                            pya.Point(x1 - ext_left, y1 - ext_lower),
                            pya.Point(x1 + ext_right, y1 + ext_upper)
                        )
                        shapes[l].insert(enc)


def _is_virtual_node_fn(n) -> bool:
    """
    Check if the node is virtual and has no direct physical representation.
    :param n:
    :return:
    """
    return n[0].startswith('virtual')


def _is_virtual_edge_fn(e) -> bool:
    """
    Check if the edge connects to at least one virtual node.
    :param n:
    :return:
    """
    a, b = e
    return _is_virtual_node_fn(a) or _is_virtual_node_fn(b)


class LcLayout:

    def __init__(self,
                 tech,
                 layout: pya.Layout,
                 placer: TransistorPlacer,
                 router: GraphRouter,
                 debug_routing_graph: bool = False,
                 debug_smt_solver: bool = False
                 ):
        assert isinstance(layout, pya.Layout)
        assert isinstance(placer, TransistorPlacer)
        assert isinstance(router, GraphRouter)

        self.tech = tech
        self.layout = layout
        self.placer = placer
        self.router = router
        self.debug_routing_graph = debug_routing_graph
        self.debug_smt_solver = debug_smt_solver

        self.cell_name = None
        self.io_pins = None
        self.SUPPLY_VOLTAGE_NET = None
        self.GND_NET = None

        # Top layout cell.
        self.top_cell: pya.Cell = None

        self._transistors_abstract: List[Transistor] = None

        self._transistor_layouts: Dict[Transistor, TransistorLayout] = None

        self.shapes: Dict[str, db.Shapes] = dict()

        self._routing_terminal_debug_layers = None

        self._abstract_cell: Cell = None
        self._cell_width = None
        self._cell_height = None

        self._spacing_graph = None

        # Routing graph.
        self._routing_graph: nx.Graph = None

        self._routing_trees = None

        # Pin definitions.
        self._pin_shapes = defaultdict(list)

    def _00_00_check_tech(self):
        # Assert that the layers in the keys of multi_via are ordered.
        for l1, l2 in self.tech.multi_via.keys():
            assert l1 <= l2, Exception('Layers must be ordered alphabetically. (%s <= %s)' % (l1, l2))

    def _00_01_prepare_tech(self):
        # Load spacing rules in form of a graph.
        self._spacing_graph = tech_util.spacing_graph(self.tech.min_spacing)

    def _01_load_netlist(self, netlist_path: str, cell_name: str):
        # Load netlist of cell.

        logger.info(f'Load netlist: {netlist_path}')

        self.cell_name = cell_name

        self._transistors_abstract, cell_pins = load_transistor_netlist(netlist_path, cell_name)
        self.io_pins = net_util.get_io_pins(cell_pins)

        ground_nets = {p for p in cell_pins if is_ground_net(p)}
        supply_nets = {p for p in cell_pins if is_supply_net(p)}

        assert len(ground_nets) > 0, "Could not find net name of ground."
        assert len(supply_nets) > 0, "Could not find net name of supply voltage."
        assert len(ground_nets) == 1, "Multiple ground net names: {}".format(ground_nets)
        assert len(supply_nets) == 1, "Multiple supply net names: {}".format(supply_nets)

        self.SUPPLY_VOLTAGE_NET = supply_nets.pop()
        self.GND_NET = ground_nets.pop()

        logger.info("Supply net: {}".format(self.SUPPLY_VOLTAGE_NET))
        logger.info("Ground net: {}".format(self.GND_NET))

        # Convert transistor dimensions into data base units.
        for t in self._transistors_abstract:
            t.channel_width = t.channel_width / self.tech.db_unit

        # Size transistor widths.
        logging.debug('Rescale transistors.')
        for t in self._transistors_abstract:
            t.channel_width = t.channel_width * self.tech.transistor_channel_width_sizing

            min_size = self.tech.minimum_gate_width_nfet if t.channel_type == ChannelType.NMOS else self.tech.minimum_gate_width_pfet

            if t.channel_width + 1e-12 < min_size:
                logger.warning("Channel width too small changing it to minimal size: %.2e < %.2e", t.channel_width,
                               min_size)

            t.channel_width = max(min_size, t.channel_width)

    def _02_setup_layout(self):
        logger.debug("Setup layout.")
        self.top_cell = self.layout.create_cell(self.cell_name)

        # Setup layers.
        self.shapes = dict()
        for name, (num, purpose) in layermap.items():
            layer = self.layout.layer(num, purpose)
            self.shapes[name] = self.top_cell.shapes(layer)

        if self.debug_routing_graph:
            # Layers for displaying routing terminals.
            self._routing_terminal_debug_layers = {
                l: self.layout.layer(idx, 200) for l, (idx, _) in layermap.items()
            }

    def _03_place_transistors(self):
        # Place transistors
        logging.info('Find transistor placement')

        abstract_cell = self.placer.place(self._transistors_abstract)
        logger.info(f"Cell placement:\n\n{abstract_cell}\n")

        self._abstract_cell = abstract_cell

    def _04_draw_transistors(self):
        logger.debug("Draw transistors.")
        # Get the locations of the transistors.
        transistor_locations = self._abstract_cell.get_transistor_locations()

        # Create the layouts of the single transistors. Layouts are already translated to the absolute position.
        self._transistor_layouts = {t: DefaultTransistorLayout(t, (x, y), self.tech)
                                    for t, (x, y) in transistor_locations}

        # Draw the transistors
        for l in self._transistor_layouts.values():
            assert isinstance(l, TransistorLayout)
            l.draw(self.shapes)

    def _05_draw_cell_template(self):
        logger.debug("Draw cell template.")

        tech = self.tech

        # Calculate dimensions of cell.
        num_unit_cells = self._abstract_cell.width
        self._cell_width = (num_unit_cells + 1) * tech.unit_cell_width
        self._cell_height = tech.unit_cell_height

        # Draw cell template.
        cell_template.draw_cell_template(self.shapes,
                                         cell_shape=(self._cell_width, self._cell_height),
                                         nwell_pwell_spacing=self._spacing_graph[l_nwell][l_pwell]['min_spacing']
                                         )

        # Draw power rails.
        vdd_rail = pya.Path([pya.Point(0, tech.unit_cell_height), pya.Point(self._cell_width, tech.unit_cell_height)],
                            tech.power_rail_width)
        vss_rail = pya.Path([pya.Point(0, 0), pya.Point(self._cell_width, 0)], tech.power_rail_width)

        # Insert power rails into layout.
        self.shapes[tech.power_layer].insert(vdd_rail).set_property('net', self.SUPPLY_VOLTAGE_NET)
        self.shapes[tech.power_layer].insert(vss_rail).set_property('net', self.GND_NET)

        # Add pin shapes for power rails.
        self.shapes[tech.pin_layer + '_pin'].insert(vdd_rail)
        self.shapes[tech.pin_layer + '_pin'].insert(vss_rail)

        # Register Pins/Ports for LEF file.
        self._pin_shapes[self.SUPPLY_VOLTAGE_NET].append((tech.power_layer, vdd_rail))
        self._pin_shapes[self.GND_NET].append((tech.power_layer, vss_rail))

    def _06_route(self):

        # TODO: Move as much as possible of the grid construction into a router specific class.
        tech = self.tech

        # Construct two dimensional grid which defines the routing graph on a single layer.
        grid = Grid2D((tech.grid_offset_x, tech.grid_offset_y),
                      (
                          tech.grid_offset_x + self._cell_width - tech.grid_offset_x,
                          tech.grid_offset_y + tech.unit_cell_height),
                      (tech.routing_grid_pitch_x, tech.routing_grid_pitch_y))

        # Create base graph
        G = create_routing_graph_base(grid, tech)

        # Remove illegal routing nodes from graph and get a dict of legal routing nodes per layer.
        remove_illegal_routing_edges(G, self.shapes, tech)

        # if not debug_routing_graph:
        #     assert nx.is_connected(G)

        # Remove pre-routed edges from G.
        remove_existing_routing_edges(G, self.shapes, tech)

        # Create a list of terminal areas: [(net, layer, [terminal, ...]), ...]
        terminals_by_net = extract_terminal_nodes(G, self.shapes, tech)

        # Embed transistor terminal nodes in to routing graph.
        embed_transistor_terminal_nodes(G, terminals_by_net, self._transistor_layouts, tech)

        # Remove terminals of nets with only one terminal. They need not be routed.
        # This can happen if a net is already connected by abutment of two transistors.
        # Count terminals of a net.
        num_appearance = Counter(chain((net for net, _, _ in terminals_by_net), self.io_pins))
        terminals_by_net = [t for t in terminals_by_net if num_appearance[t[0]] > 1]

        # Check if each net really has a routing terminal.
        # It can happen that there is none due to spacing issues.
        # First find all net names in the layout.
        all_net_names = {s.property('net') for _, _shapes in self.shapes.items() for s in _shapes.each()}
        all_net_names -= {None}

        error = False
        # Check if each net has at least a terminal.
        for net_name in all_net_names:
            num_terminals = num_appearance.get(net_name)
            if num_terminals is None or num_terminals == 0:
                logger.error("Net '{}' has no routing terminal.".format(net_name))
                error = True

        if not self.debug_routing_graph:
            assert not error, "Nets without terminals. Check the routing graph (--debug-routing-graph)!"

        # Create virtual graph nodes for each net terminal.
        virtual_terminal_nodes = create_virtual_terminal_nodes(G, terminals_by_net, self.io_pins, tech)

        if self.debug_routing_graph:
            # Display terminals on layout.
            routing_terminal_shapes = {
                l: self.top_cell.shapes(self._routing_terminal_debug_layers[l]) for l in tech.routing_layers.keys()
            }
            for net, layer, ts in terminals_by_net:
                for x, y in ts:
                    d = tech.routing_grid_pitch_x // 16
                    routing_terminal_shapes[layer].insert(pya.Box(pya.Point(x - d, y - d), pya.Point(x + d, y + d)))

        # Remove nodes that will not be used for routing.
        # Iteratively remove nodes of degree 1.
        while True:
            unused_nodes = set()
            for n in G:
                if nx.degree(G, n) <= 1:
                    if not _is_virtual_node_fn(n):
                        unused_nodes.add(n)
            if len(unused_nodes) == 0:
                break
            G.remove_nodes_from(unused_nodes)

        if not nx.is_connected(G):
            assert False, 'Routing graph is not connected.'

        self._routing_graph = G

        # TODO: SPLIT HERE
        # def _07_route(self):

        tech = self.tech
        spacing_graph = self._spacing_graph
        G = self._routing_graph

        # Route
        if self.debug_routing_graph:
            # Write the full routing graph to GDS.
            logger.info("Skip routing and plot routing graph.")
            self._routing_trees = {'graph': self._routing_graph}
        else:
            logger.info("Start routing")
            # For each routing node find other nodes that are close enough that they cannot be used
            # both for routing. This is used to avoid spacing violations during routing.
            logger.debug("Find conflicting nodes.")
            conflicts = dict()
            # Loop through all nodes in the routing graph G.
            for n in G:
                # Skip virtual nodes which have no physical representation.
                if not _is_virtual_node_fn(n):
                    layer, point = n
                    wire_width1 = tech.wire_width.get(layer, 0) // 2
                    node_conflicts = set()
                    if layer in spacing_graph:
                        # If there is a spacing rule defined involving `layer` then
                        # loop through all layers that have a spacing rule defined
                        # relative to the layer of the current node n.
                        for other_layer in spacing_graph[layer]:
                            if other_layer in tech.routing_layers:
                                # Find minimal spacing of nodes such that spacing rule is asserted.
                                wire_width2 = tech.wire_width.get(other_layer, 0) // 2
                                min_spacing = spacing_graph[layer][other_layer]['min_spacing']
                                margin = (wire_width1 + wire_width2 + min_spacing)

                                # Find nodes that are closer than the minimal spacing.
                                # conflict_points = grid.neigborhood(point, margin, norm_ord=1)
                                potential_conflicts = [x for x in G if x[0] == other_layer]
                                conflict_points = [p for (_, p) in potential_conflicts
                                                   if numpy.linalg.norm(numpy.array(p) - numpy.array(point),
                                                                        ord=1) < margin
                                                   ]
                                # Construct the lookup table for conflicting nodes.
                                for p in conflict_points:
                                    conflict_node = other_layer, p
                                    if conflict_node in G:
                                        node_conflicts.add(conflict_node)
                    if node_conflicts:
                        conflicts[n] = node_conflicts

            # Find routing nodes that are reserved for a net. They cannot be used to route other nets.
            # (For instance the ends of a gate stripe.)
            reserved_nodes = defaultdict(set)
            for net, layer, terminals in terminals_by_net:
                for p in terminals:
                    n = layer, p
                    reserved = reserved_nodes[net]
                    reserved.add(n)
                    if n in conflicts:
                        for c in conflicts[n]:  # Also reserve nodes that would cause a spacing violation.
                            reserved.add(c)

            assert nx.is_connected(G)

            # Invoke router and store result.
            self._routing_trees = self.router.route(G,
                                                    signals=virtual_terminal_nodes,
                                                    reserved_nodes=reserved_nodes,
                                                    node_conflict=conflicts,
                                                    is_virtual_node_fn=_is_virtual_node_fn
                                                    )

            # TODO: Sanity check on result.

    def _08_draw_routes(self):
        # Draw the layout of the routes.
        for signal_name, rt in self._routing_trees.items():
            _draw_routing_tree(self.shapes, self._routing_graph, rt, self.tech, self.debug_routing_graph)

        # Merge the polygons on all layers.
        _merge_all_layers(self.shapes)

    def _09_post_process(self):
        tech = self.tech
        # Register Pins/Ports for LEF file.

        if not self.debug_routing_graph:

            # Clean DRC violations that are not handled above.

            def fill_all_notches():
                # Remove notches on all layers.
                for layer, s in self.shapes.items():
                    if layer in tech.minimum_notch:

                        if layer in tech.connectable_layers:
                            r = pya.Region(s)
                            filled = fill_notches(r, tech.minimum_notch[layer])
                            s.insert(filled)
                        else:
                            # Remove notches per polygon to avoid connecting independent shapes.
                            s_filled = pya.Shapes()
                            for shape in s.each():
                                r = pya.Region(shape.polygon)

                                filled = fill_notches(r, tech.minimum_notch[layer])
                                s_filled.insert(filled)

                            s.insert(s_filled)

                    _merge_all_layers(self.shapes)

            # Fill notches that violate a notch rule.
            fill_all_notches()
            # Do a second time because first iteration could have introduced new notch violations.
            fill_all_notches()

            # Fix minimum area violations.
            fix_min_area(tech, self.shapes, debug=self.debug_smt_solver)

            # Draw pins
            # Get shapes of pins.
            pin_locations_by_net = {}
            pin_shapes_by_net = {}
            for net_name, rt in self._routing_trees.items():
                # Get virtual pin nodes.
                virtual_pins = [n for n in rt.nodes if n[0] == 'virtual_pin']
                for vp in virtual_pins:
                    # Get routing nodes adjacent to virtual pin nodes. They contain the location of the pin.
                    locations = [l for _, l in rt.edges(vp)]
                    _, net_name, _, _ = vp
                    for layer, (x, y) in locations:
                        w = tech.minimum_pin_width
                        s = self.shapes[layer]

                        # Find shape at (x,y).
                        ball = pya.Box(pya.Point(x - 1, y - 1), pya.Point(x + 1, y + 1))
                        pin_shapes = pya.Region(s).interacting(pya.Region(ball))

                        # Remember pin location and shape.
                        pin_locations_by_net[net_name] = x, y
                        pin_shapes_by_net[net_name] = pin_shapes

                        # Register pin shapes for LEF file.
                        self._pin_shapes[net_name].append((layer, pin_shapes))

            # Add pin labels
            for net_name, (x, y) in pin_locations_by_net.items():
                logger.debug('Add pin label: %s, (%d, %d)', net_name, x, y)
                _draw_label(self.shapes, tech.pin_layer + '_label', (x, y), net_name)

            # Add label for power rails
            _draw_label(self.shapes, tech.power_layer + '_label', (self._cell_width // 2, 0), self.GND_NET)
            _draw_label(self.shapes, tech.power_layer + '_label', (self._cell_width // 2, self._cell_height),
                        self.SUPPLY_VOLTAGE_NET)

            # Add pin shapes.
            for net_name, pin_shapes in pin_shapes_by_net.items():
                self.shapes[tech.pin_layer + '_pin'].insert(pin_shapes)

    def create_cell_layout(self, cell_name: str, netlist_path: str) \
            -> Tuple[pya.Cell, Dict[str, List[Tuple[str, pya.Shape]]]]:

        self._00_00_check_tech()
        self._00_01_prepare_tech()

        self._01_load_netlist(netlist_path, cell_name)

        self._02_setup_layout()
        self._03_place_transistors()
        self._04_draw_transistors()
        self._05_draw_cell_template()
        self._06_route()
        self._08_draw_routes()
        self._09_post_process()

        return self.top_cell, self._pin_shapes


def main():
    """
    Entry function for standalone command line tool.
    :return:
    """
    import argparse
    import datetime
    import time

    # List of available placer engines.
    placers = {
        'flat': EulerPlacer,
        'hierarchical': HierarchicalPlacer,
        'smt': SMTPlacer
    }

    signal_routers = {
        'dijkstra': DijkstraRouter,  # Fast but not stable.
        'steiner': ApproxSteinerTreeRouter,  # Slow but best results.
        # 'lp': LPSignalRouter
    }

    # Define commandline arguments.
    parser = argparse.ArgumentParser(description='Generate GDS layout from SPICE netlist.')
    parser.add_argument('--cell', required=True, metavar='NAME', type=str, help='cell name')
    parser.add_argument('--netlist', required=True, metavar='FILE', type=str, help='path to SPICE netlist')
    parser.add_argument('--output-dir', default='.', metavar='DIR', type=str, help='output directory for layouts')
    parser.add_argument('--tech', required=True, metavar='FILE', type=str, help='technology file')

    parser.add_argument('--debug-routing-graph', action='store_true',
                        help='write full routing graph to the layout instead of wires')
    parser.add_argument('--debug-smt-solver', action='store_true',
                        help='enable debug mode: display routing nodes in layout, \
                        show unsatisfiable core if SMT DRC cleaning fails.')

    parser.add_argument('--placer', default='flat', metavar='PLACER', type=str, choices=placers.keys(),
                        help='placement algorithm ({})'.format(', '.join(sorted(placers.keys()))))

    parser.add_argument('--signal-router', default='dijkstra', metavar='SIGNAL_ROUTER', type=str,
                        choices=signal_routers.keys(),
                        help='routing algorithm for single signals ({})'.format(
                            ', '.join(sorted(signal_routers.keys()))))

    # parser.add_argument('--profile', action='store_true', help='enable profiler')
    parser.add_argument('-v', '--verbose', action='store_true', help='show more information')
    parser.add_argument('--ignore-lvs', action='store_true', help='Write the layout file even if the LVS check failed.')
    parser.add_argument('-q', '--quiet', action='store_true',
                        help="don't show any information except fatal events (overwrites --verbose)")
    parser.add_argument('--log', required=False, metavar='LOG_FILE', type=str,
                        help='write log to this file instead of stdout')

    # Parse arguments
    args = parser.parse_args()

    log_level = logging.INFO
    if args.verbose:
        log_level = logging.DEBUG
    if args.quiet:
        log_level = logging.FATAL

    # Setup logging
    logging.basicConfig(format='%(asctime)s %(module)16s %(levelname)8s: %(message)s',
                        datefmt="%Y-%m-%d %H:%M:%S",
                        level=log_level,
                        filename=args.log)

    # Load netlist of cell
    cell_name = args.cell
    netlist_path = args.netlist

    tech_file = args.tech

    tech = tech_util.load_tech_file(tech_file)

    # Create empty layout
    layout = pya.Layout()

    # Setup placer algorithm

    placer = placers[args.placer]()
    logger.info("Placement algorithm: {}".format(type(placer).__name__))

    # Setup routing algorithm

    signal_router = signal_routers[args.signal_router]()
    logger.info("Signal routing algorithm: {}".format(type(signal_router).__name__))

    router = PathFinderGraphRouter(
        signal_router
    )
    # router = LPGraphRouter()
    router = HVGraphRouter(router,
                           orientation_change_penalty=tech.orientation_change_penalty
                           )

    layouter = LcLayout(tech=tech,
                        layout=layout,
                        placer=placer,
                        router=router,
                        debug_routing_graph=args.debug_routing_graph,
                        debug_smt_solver=args.debug_smt_solver)

    # Run layout synthesis
    time_start = time.process_time()
    cell, pin_geometries = layouter.create_cell_layout(cell_name, netlist_path)

    # LVS check
    logger.info("Running LVS check")
    reference_netlist = lvs.read_netlist_mos4_to_mos3(netlist_path)

    # Remove all unused circuits.
    # The reference netlist must contain only the circuit of the cell to be checked.
    # Copying a circuit into a new netlist makes `combine_devices` fail.
    circuits_to_delete = {c for c in reference_netlist.each_circuit() if c.name != cell_name}
    for c in circuits_to_delete:
        reference_netlist.remove(c)

    # Extract netlist from layout.
    extracted_netlist = lvs.extract_netlist(layout, cell)

    # Run LVS comparison of the two netlists.
    lvs_success = lvs.compare_netlist(extracted_netlist, reference_netlist)

    logger.info("LVS result: {}".format('SUCCESS' if lvs_success else 'FAILED'))

    if not lvs_success:
        logger.error("LVS check failed!")
        if not args.ignore_lvs and not args.debug_routing_graph:
            exit(1)

    # Output using defined output writers.
    from .writer.writer import Writer
    for writer in tech.output_writers:
        assert isinstance(writer, Writer)
        logger.debug("Call output writer: {}".format(type(writer).__name__))
        writer.write_layout(
            layout=layout,
            pin_geometries=pin_geometries,
            top_cell=cell,
            output_dir=args.output_dir
        )

    time_end = time.process_time()
    duration = datetime.timedelta(seconds=time_end - time_start)
    logger.info("Done (Total duration: {})".format(duration))


def fix_min_area(tech, shapes: Dict[str, pya.Shapes], debug=False):
    """
    Fix minimum area violations.
    This is a wrapper around the drc_cleaner module.
    :param tech:
    :param shapes:
    :param debug: Tell DRC cleaner to find unsatisiable core.
    :return:
    """

    # Find minimum area violations.
    # And create a set of whitelisted shapes that are allowed to be changed for DRC cleaning.
    min_area_violations = set()
    for layer, _shapes in shapes.items():
        min_area = tech.min_area.get(layer, 0)
        for shape in _shapes.each():
            area = shape.area()
            if area < min_area:
                min_area_violations.add((layer, shape))

    # TODO: Also whitelist vias connected to the violating shapes.

    if min_area_violations:
        success = drc_cleaner.clean(tech,
                                    shapes=shapes,
                                    white_list=min_area_violations,
                                    enable_min_area=True,
                                    debug=debug
                                    )
        if not success:
            logger.error("Minimum area fixing failed!")
    else:
        logger.info("No minimum area violations.")



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from .signal_router import SignalRouter
import networkx as nx
from typing import Any, Dict, Set, List
from itertools import product


class MinAreaRouter(SignalRouter):
    """ Wrap a detail router function such that it respects minimum area constraints.

    Parameters
    ----------

    router: Underlying detail router
    node_conflict: Dict[Node, Set[Node]]
            A mapping from a node `n` to a set of nodes that collide with `n`.
    """

    def __init__(self, router: SignalRouter, node_conflict: Dict[Any, Set], min_area: Dict[Any, float]):
        self.router = router
        self.node_conflict = node_conflict
        self.min_area = min_area

    def route(self, G: nx.Graph,
              terminals: List[List[Any]],
              node_cost_fn,
              edge_cost_fn
              ) -> nx.Graph:
        return self._min_area_route(self.router,
                                    G,
                                    terminals,
                                    self.node_conflict,
                                    node_cost_fn,
                                    edge_cost_fn)

    def _min_area_route(self, router: SignalRouter, G: nx.Graph, terminals, node_conflict: Dict, node_cost_fn,
                        edge_cost_fn):
        S1 = router.route(G, terminals, node_cost_fn, edge_cost_fn)

        nodes = list(S1.nodes)

        # Get layers for which min_area is defined.
        min_area_layers = set(self.min_area.keys())

        return S1





#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import networkx as nx
from itertools import count
from collections import defaultdict

from .layout.grid_helpers import *
from .layout.geometry_helpers import *
from .layout.grid import Grid2D
from .layout.layers import *
from .layout.transistor import TransistorLayout
from .data_types import Transistor
from . import tech_util

from typing import Any, Dict, List, Tuple, Iterable, Set
import logging

logger = logging.getLogger(__name__)


def create_routing_graph_base(grid: Grid2D, tech) -> nx.Graph:
    """ Construct the full mesh of the routing graph.
    :param grid: grid points
    :param tech: module containing technology information
    :return: nx.Graph
    """
    logging.debug('Create routing graph.')

    # Create routing graph.

    # Create nodes and vias.
    G = nx.Graph()

    # Create nodes on routing layers.
    for layer, directions in tech.routing_layers.items():
        for p in grid:
            n = layer, p
            G.add_node(n)

    # Create via edges.
    for l1, l2, data in via_layers.edges(data=True):
        via_layer = data['layer']
        for p in grid:
            n1 = (l1, p)
            n2 = (l2, p)

            weight = tech.via_weights.get((l1, l2))
            if weight is None:
                weight = tech.via_weights[(l2, l1)]

            multi_via = tech.multi_via.get((l1, l2))
            if multi_via is None:
                multi_via = tech.multi_via.get((l2, l1), 1)

            # Create edge: n1 -- n2
            G.add_edge(n1, n2,
                       weight=weight,
                       multi_via=multi_via,
                       layer=via_layer
                       )

    # Create intra layer routing edges.
    for layer, directions in tech.routing_layers.items():
        for p1 in grid:
            x1, y1 = p1
            x2 = x1 + tech.routing_grid_pitch_x
            y2 = y1 + tech.routing_grid_pitch_y

            # ID of the graph node.
            n = layer, p1

            # Horizontal edge.
            if 'h' in directions:
                n_right = layer, (x2, y1)
                if n_right in G.nodes:
                    weight = tech.weights_horizontal[layer] * abs(x2 - x1)
                    G.add_edge(n, n_right, weight=weight, orientation='h', layer=layer)

            # Vertical edge.
            if 'v' in directions:
                n_upper = layer, (x1, y2)
                if n_upper in G.nodes:
                    weight = tech.weights_vertical[layer] * abs(y2 - y1)
                    G.add_edge(n, n_upper, weight=weight, orientation='v', layer=layer)

    assert nx.is_connected(G)
    return G


def _get_routing_node_locations_per_layer(g: nx.Graph) -> Dict[Any, Set[Tuple[int, int]]]:
    """ For each layer extract the positions of the routing nodes.

    :param g: Routing graph.
    :return: Dict[layer name, set of (x,y) coordinates of routing nodes]
    """
    # Dict that will contain for each layer the node coordinates that can be used for routing.
    routing_nodes = defaultdict(set)
    # Populate `routing_nodes`
    for e in g.edges:
        (l1, p1), (l2, p2) = e
        routing_nodes[l1].add(p1)
        routing_nodes[l2].add(p2)

    return routing_nodes


def remove_illegal_routing_edges(graph: nx.Graph, shapes: Dict[Any, pya.Shapes], tech) -> None:
    """ Remove nodes and edges from  G that would conflict
    with predefined `shapes`.
    :param graph: routing graph.
    :param shapes: Dict[layer name, pya.Shapes]
    :param tech: module containing technology information
    :return: Dict[layer name, List[Node]]
    """

    # Build a spacing rule graph by mapping the minimal spacing between layer a and layer b to an edge
    # a-b in the graph with weight=min_spacing.
    spacing_graph = tech_util.spacing_graph(tech.min_spacing)

    # Get a dict mapping layer names to pya.Regions
    regions = {l: pya.Region(s) for l, s in shapes.items()}
    illegal_edges = set()
    # For each edge in the graph check if it conflicts with an existing shape.
    # Remember the edge if it is in conflict.
    for e in graph.edges:
        (l1, p1), (l2, p2) = e
        is_via = l1 != l2

        if not is_via:
            layer = l1
            other_layers = spacing_graph[layer]
            for other_layer in other_layers:
                if other_layer != layer:
                    min_spacing = spacing_graph[layer][other_layer]['min_spacing']
                    wire_width_half = (tech.wire_width.get(layer, 0) + 1) // 2
                    margin = wire_width_half + min_spacing
                    # TODO: treat horizontal and vertical lines differently if they don't have the same wire width.
                    region = regions[other_layer]
                    is_illegal_edge = interacts(p1, region, margin) or interacts(p2, region, margin)

                    if is_illegal_edge:
                        illegal_edges.add(e)
        else:
            assert p1 == p2, "End point coordinates of a via edge must match."
            layer = via_layers[l1][l2]['layer']
            if layer in spacing_graph:
                other_layers = spacing_graph[layer]
                for other_layer in other_layers:
                    if other_layer != layer:
                        if layer in spacing_graph and other_layer in spacing_graph:
                            min_spacing = spacing_graph[layer][other_layer]['min_spacing']
                            via_width_half = (tech.via_size[layer] + 1) // 2
                            margin = via_width_half + min_spacing
                            region = regions[other_layer]
                            is_illegal_edge = interacts(p1, region, margin)

                            if is_illegal_edge:
                                illegal_edges.add(e)

    # Now remove all edges from G that are in conflict with existing shapes.
    graph.remove_edges_from(illegal_edges)

    # Remove unconnected nodes.
    unconnected = set()
    for n in graph:
        d = nx.degree(graph, n)
        if d < 1:
            unconnected.add(n)
    graph.remove_nodes_from(unconnected)


def remove_existing_routing_edges(G: nx.Graph, shapes: Dict[Any, pya.Shapes], tech) -> None:
    """ Remove edges in G that are already routed by a shape in `shapes`.
    :param G: Routing graph to be modified.
    :param shapes: Dict[layer, pya.Shapes]
    :param tech: module containing technology information
    :return: None
    """

    # Remove all routing edges that are inside existing shapes.
    # (They are already connected and cannot be used for routing).
    for l in tech.routing_layers.keys():
        edges = edges_inside(G, pya.Region(shapes[l]), 1)
        for e in edges:
            (l1, _), (l2, _) = e
            if (l1, l2) == (l, l):
                G.remove_edge(*e)


def extract_terminal_nodes(graph: nx.Graph,
                           shapes: Dict[str, pya.Shapes],
                           tech):
    """ Get terminal nodes for each net.
    :param graph: Routing graph.
    :param net_regions: Regions that are connected to a net: Dict[net, Dict[layer, pya.Region]]
    :param tech: module containing technology information
    :return: list of terminals: [(net, layer, [terminal, ...]), ...]
    """

    routing_nodes = _get_routing_node_locations_per_layer(graph)

    # Create a list of terminal areas: [(net, layer, [terminal, ...]), ...]
    terminals_by_net = []
    for layer, _shapes in shapes.items():
        for net_shape in _shapes.each():
            net = net_shape.property('net')

            if net is not None:
                possible_via_layers = [data['layer'] for _, _, data in via_layers.edges(layer, data=True)]
                enc = max((tech.minimum_enclosure.get((layer, via_layer), 0) for via_layer in possible_via_layers))
                max_via_size = max((tech.via_size[l] for l in possible_via_layers))

                # TODO: How to convert db.Shape into db.Region in a clean way???

                s = db.Shapes()
                s.insert(net_shape)
                terminal_region = pya.Region(s)

                if layer in tech.routing_layers:
                    # On routing layers enclosure can be added, so nodes are not required to be properly enclosed.
                    d = 1
                    routing_terminals = interacting(routing_nodes[layer], terminal_region, d)
                else:
                    # A routing node must be properly enclosed to be used.
                    d = enc + max_via_size // 2
                    routing_terminals = inside(routing_nodes[layer], terminal_region, d)

                terminals_by_net.append((net, layer, routing_terminals))
                # Don't use terminals for normal routing
                routing_nodes[layer] -= set(routing_terminals)
                # TODO: need to be removed from G also. Better: construct edges in G afterwards.

    # for net, regions in net_regions.items():
    #     for layer, region in regions.items():
    #         if layer in routing_nodes:
    #             for net_shape in region.each_merged():
    #
    #                 possible_via_layers = [data['layer'] for _, _, data in via_layers.edges(layer, data=True)]
    #                 enc = max((tech.minimum_enclosure.get((layer, via_layer), 0) for via_layer in possible_via_layers))
    #                 max_via_size = max((tech.via_size[l] for l in possible_via_layers))
    #
    #                 if layer in tech.routing_layers:
    #                     # On routing layers enclosure can be added, so nodes are not required to be properly enclosed.
    #                     d = 1
    #                     routing_terminals = interacting(routing_nodes[layer], pya.Region(net_shape), d)
    #                 else:
    #                     # A routing node must be properly enclosed to be used.
    #                     d = enc + max_via_size // 2
    #                     routing_terminals = inside(routing_nodes[layer], pya.Region(net_shape), d)
    #
    #                 terminals_by_net.append((net, layer, routing_terminals))
    #                 # Don't use terminals for normal routing
    #                 routing_nodes[layer] -= set(routing_terminals)
    #                 # TODO: need to be removed from G also. Better: construct edges in G afterwards.
    #         else:
    #             logger.warning("Layer '{}' does not contain any routing nodes.".format(layer))

    # # Sanity check
    # error = False
    # for net_name, layer, terminals in terminals_by_net:
    #     if len(terminals) == 0:
    #         logger.error(
    #             "Shape of net {} on layer '{}' does not contain any routing grid point.".format(net_name, layer))
    #         error = True

    return terminals_by_net


def embed_transistor_terminal_nodes(G: nx.Graph,
                                    terminals_by_net: List[Tuple[str, str, List[Tuple[int, int]]]],
                                    transistor_layouts: Dict[Transistor, TransistorLayout],
                                    tech):
    """ Embed the terminal nodes of a the transistors into the routing graph.
    Modifies `G` and `terminals_by_net`
    :param G: The routing graph.
    :param terminals_by_net:
    :param transistor_layouts: List[TransistorLayout]
    :param tech: module containing technology information
    :return: None
    """
    # Connect terminal nodes of transistor gates in G.
    for t_layout in transistor_layouts.values():
        terminals = t_layout.terminal_nodes()
        for net, ts in terminals.items():
            for t in ts:
                layer, (x, y) = t

                logger.info(f"Terminal node {net} {layer} {t}")

                # Insert terminal into G.
                next_x = grid_round(x, tech.routing_grid_pitch_x, tech.grid_offset_x)

                assert next_x == x, Exception("Terminal node not x-aligned.")

                x_aligned_nodes = [(l, (_x, y)) for l, (_x, y) in G if l == layer and _x == x]

                def dist(a, b):
                    _, (x1, y1) = a
                    _, (x2, y2) = b
                    return (x1 - x2) ** 2 + (y1 - y2) ** 2

                neighbour_node = min(x_aligned_nodes, key=lambda n: dist(n, t))

                # TODO: weight proportional to gate width?
                G.add_edge(t, neighbour_node, weight=1000, wire_width=tech.gate_length)

            coords = [c for _, c in ts]
            terminals_by_net.append((net, layer, coords))


def create_virtual_terminal_nodes(G: nx.Graph,
                                  terminals_by_net: List[Tuple[str, str, Tuple[int, int]]],
                                  io_pins: Iterable,
                                  tech):
    """ Create virtual terminal nodes for each net.
    :param G: The routing graph. Will be modified.
    :param terminals_by_net:
    :param io_pins: Names of the I/O nets.
    :param tech: module containing technology information
    :return: Returns a set of virtual terminal nodes: Dict[('virtual...', net, layer, id)]
    """

    # Extract all routing nodes for each layer.
    routing_nodes = _get_routing_node_locations_per_layer(G)

    # Create virtual graph nodes for each net terminal.
    virtual_terminal_nodes = defaultdict(list)
    cnt = count()

    for net, layer, terminals in terminals_by_net:
        weight = 1000
        if len(terminals) > 0:
            # if layer in (l_ndiffusion, l_pdiffusion) and False:  # TODO: make tech independet
            #     for p in terminals:
            #         # Force router to connect to all contacts to a l_active shape.
            #         virtual_net_terminal = ('virtual', net, layer, next(cnt))
            #         virtual_terminal_nodes[net].append(virtual_net_terminal)
            #         n = layer, p
            #         assert n in G.nodes, "Node not present in graph: %s" % str(n)
            #         G.add_edge(virtual_net_terminal, n, weight=weight)
            # else:
            virtual_net_terminal = ('virtual', net, layer, next(cnt))
            virtual_terminal_nodes[net].append(virtual_net_terminal)

            for p in terminals:
                n = layer, p
                assert n in G.nodes, "Node not present in graph: %s" % str(n)
                # High weight for virtual edge
                # TODO: High weight only for low-resistance layers.
                G.add_edge(virtual_net_terminal, n, weight=weight)

    cnt = count()
    # Create virtual nodes for I/O pins.
    for p in io_pins:
        virtual_net_terminal = ('virtual_pin', p, tech.pin_layer, next(cnt))
        virtual_terminal_nodes[p].append(virtual_net_terminal)

        for p in routing_nodes[tech.pin_layer]:
            n = tech.pin_layer, p
            x, y = p
            assert n in G.nodes, "Node not present in graph: %s" % str(n)
            # A huge weight assures that the virtual node is not used as a worm hole for routing.
            weight = 10000000 + (y - tech.unit_cell_height // 2) // tech.routing_grid_pitch_y
            G.add_edge(virtual_net_terminal, n, weight=weight)

    return virtual_terminal_nodes



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from .layers import *
from klayout import db
from typing import Dict, Tuple

"""
      # Fill half of the cell with nwell.
        # TODO: do this in the cell template or after placing the transistors.
        nwell_box = pya.Box(
            pya.Point(0, cell_height // 2),
            pya.Point(cell_width, cell_height)
        )

        shapes[l_nwell].insert(nwell_box)
"""


def draw_cell_template(shapes: Dict[str, db.Shapes],
                       cell_shape: Tuple[int, int],
                       nwell_pwell_spacing: int = 0) -> None:
    """
    Draw shapes of the cell that can be drawn without knowledge of the transistor placement.
    This includes the cell boundary, nwell/pwell.
    :param shapes: KLayout shapes that will be modified.
    :param cell_shape: Dimensions of the cell (width, height).
    :param nwell_pwell_spacing: Minimum spacing between nwell and pwell for twin-well layouts.
    :return: None
    """
    cell_width, cell_height = cell_shape

    # Draw abutment box.
    shapes[l_abutment_box].insert(db.Box(0, 0, cell_width, cell_height))

    # Draw nwell / pwell

    # Fill half of the cell with nwell.

    nwell_start_y = cell_height // 2 + nwell_pwell_spacing // 2
    nwell_end_y = cell_height

    pwell_start_y = 0
    pwell_end_y = cell_height // 2 - nwell_pwell_spacing // 2

    nwell_box = db.Box(
        db.Point(0, nwell_start_y),
        db.Point(cell_width, nwell_end_y)
    )

    shapes[l_nwell].insert(nwell_box)

    pwell_box = db.Box(
        db.Point(0, pwell_start_y),
        db.Point(cell_width, pwell_end_y)
    )
    shapes[l_pwell].insert(pwell_box)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
"""
Helper functions for finding minima and maxima.
"""

from typing import Iterable, List


def all_max(args: Iterable, key = None) -> List:
    """ Find all global maxima in `args`.
    :param args:
    :param key: Key function.
    :return: List of maxima.
    """

    if key is None:
        def key(x):
            return x

    maxima = []
    max_key = None

    for x in args:
        val = key(x)
        if max_key is None or val > max_key:
            maxima.clear()
            max_key = val
        if val == max_key:
            maxima.append(x)

    return maxima


def all_min(args: Iterable, key = None) -> List:
    """ Find all global minima in `args`.
    :param args:
    :param key: Key function.
    :return: List of minima.
    """

    if key is None:
        def key(x):
            return x

    minima = []
    min_key = None

    for x in args:
        val = key(x)
        if min_key is None or val < min_key:
            minima.clear()
            min_key = val
        if val == min_key:
            minima.append(x)

    return minima



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
from setuptools import setup, find_packages


def readme():
    with open("README.md", "r") as f:
        return f.read()


setup(name='librecell-layout',
      version='0.0.8',
      description='CMOS standard cell layout generator.',
      long_description=readme(),
      long_description_content_type="text/markdown",
      keywords='cmos cell generator layout klayout vlsi asic',
      classifiers=[
          # 'License :: OSI Approved :: GNU Affero General Public License v3',
          'Development Status :: 3 - Alpha',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
          'Programming Language :: Python :: 3'
      ],
      url='https://codeberg.org/tok/librecell',
      author='T. Kramer',
      author_email='code@tkramer.ch',
      license='OHL-S v2.0',
      packages=find_packages(),
      package_data={'': ['examples/*']},
      entry_points={
          'console_scripts': [
              'lclayout = lclayout.standalone:main',
              # 'drc_cleaner = lclayout.drc_cleaner.standalone:main',
          ]
      },
      install_requires=[
          'librecell-common==0.0.7',
          'toml==0.10.*',
          'klayout==0.26.*',  # GPLv3
          'numpy==1.*',  # BSD
          'networkx==2.5',  # BSD
          'pyspice==1.4.3',  # GPLv3
          'scipy>=1.5.*',  # BSD
          'liberty-parser==0.0.8',  # GPLv3
          'pysmt==0.9.*',  # Apache-2.0
          'z3-solver==4.8.*',  #
      ],
      zip_safe=False)



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#

import pyo3_cell
import networkx as nx

def route(G, signals, reserved_nodes = None, node_conflict=None):
	
	if isinstance(signals, dict):
		signals = list(signals.values())
	
	if isinstance(reserved_nodes, dict):
		reserved_nodes = list(reserved_nodes.values())
		
	if reserved_nodes is None:
		reserved_nodes = []*len(signals)
	
	rust_edges, nodemap = graph_networkx2rust(G)
	rust_signals = [[nodemap[t] for t in terminals] for terminals in signals]
	
	node_collisions = list(node_conflict.items()) if node_conflict else []
	node_collisions = [(nodemap[n], [nodemap[c] for c in collisions]) for n, collisions in node_collisions]
	
	reserved_nodes = [[nodemap[n] for n in nodes] for nodes in reserved_nodes]
	
	rust_routing_trees = pyo3_cell.route_mst(rust_edges, rust_signals, reserved_nodes, node_collisions)
	
	routing_trees = graph_rust2networkx(rust_routing_trees, nodemap)
	return routing_trees

def graph_networkx2rust(G):
	# Map nodes to indices
	nodemap = {n: i for i,n in enumerate(G.nodes)}
	
	rustgraph = nx.Graph()
	rustgraph.add_nodes_from((nodemap[n] for n in G.nodes))
	rustgraph.add_edges_from(((nodemap[a], nodemap[b], d) for a,b,d in G.edges(data=True)))
	
	rust_edges = list(rustgraph.edges(data=True))
	
	for a,b,data in rust_edges:
		assert data['weight'] > 0
	
	return rust_edges, nodemap
	
def graph_rust2networkx(rust_routing_trees, nodemap):

	reverse_map = {v:k for k,v in nodemap.items()}
	
	# Convert back to python nodes.
	routing_edges = [
		[(reverse_map[a], reverse_map[b]) for a,b in rt]
		for rt in rust_routing_trees
	]
	routing_trees = []
	for re in routing_edges:
		rt = nx.Graph()
		rt.add_edges_from(re)
		routing_trees.append(rt)
		
	return routing_trees



#
# Copyright 2019-2020 Thomas Kramer.
#
# This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
#
# You may redistribute and modify this documentation and make products using it
# under the terms of the CERN-OHL-S v2 (https:/cern.ch/cern-ohl).
# This documentation is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
# INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Please see the CERN-OHL-S v2 for applicable conditions.
#
# Source location: https://codeberg.org/tok/librecell
#
import klayout.db as pya

import argparse
import logging

from .. import tech_util

from .drc_cleaner import clean

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(description='DRC cleaning.')
    parser.add_argument('-i', '--input', required=True, metavar='GDS', type=str, help='GDS input file')
    parser.add_argument('-o', '--output', required=True, metavar='GDS', type=str, help='GDS output file')
    parser.add_argument('--debug', action='store_true', help='Enable debug mode.')
    parser.add_argument('--tech', required=True, metavar='FILE', type=str, help='technology file')
    parser.add_argument('--solver', default='z3', metavar='SOLVER', type=str, help='SMT solver name')
    parser.add_argument('--optimize', default=False, action='store_true', help='Enable area optimizations')

    # Parse arguments
    args = parser.parse_args()

    DEBUG = args.debug

    log_level = logging.DEBUG if DEBUG else logging.INFO

    logging.basicConfig(format='%(module)16s %(levelname)8s: %(message)s', level=log_level)

    layout = pya.Layout()
    logger.info("Reading GDS: %s", args.input)
    layout.read(args.input)
    top = layout.top_cell()

    tech = tech_util.load_tech_file(args.tech)

    # Fetch layers and create `shapes` object.
    # shapes: Dict[layer name, pya.Shapes]
    shapes = {}
    for name, (num, purpose) in tech.layermap.items():
        layer = layout.find_layer(num, purpose)
        # assert layer is not None, Exception('Layer not found: %s (%d, %d)' % (name, num, purpose))
        if layer is not None:
            shapes[name] = top.shapes(layer)

    logger.debug("shapes = %s", shapes)

    # Select all shapes that are completely inside `outline` on whitelisted layers
    # to be modified. (For instance excludes power rails from being modified.)
    white_list = set()
    outline = pya.Region(shapes['outline'])
    for layer in ['m1', 'm2', 'v1']:
        r = pya.Region(shapes[layer])
        inside = r.inside(outline)
        for shape in inside.each():
            white_list.add((layer, shape))

    drc_clean_success = clean(tech,
                              shapes,
                              white_list=white_list,
                              debug=DEBUG,
                              solver_name=args.solver,
                              optimize=args.optimize)

    if drc_clean_success:
        # Store result
        logger.info('Write GDS: %s', args.output)
        layout.write(args.output)
    else:
        logger.error("Could not clean design. GDS left unmodified.")
